from decimal import Decimal as D

from django.core.exceptions import ObjectDoesNotExist


class Scale(object):
    """
    For calculating the weight of a product or basket
    """

    def __init__(self, attribute_code="weight", default_weight=None):
        self.attribute = attribute_code
        self.default_weight = default_weight

    def weigh_product(self, product):
        weight = None
        try:
            weight = (
                product.get_attribute_values().get(attribute__code=self.attribute).value
            )
        except ObjectDoesNotExist:
            pass

        if weight is None:
            if self.default_weight is None:
                raise ValueError(
                    "No attribute %s found for product %s" % (self.attribute, product)
                )
            weight = self.default_weight

        return D(weight) if weight is not None else D("0.0")

    def weigh_basket(self, basket):
        weight = D("0.0")
        for line in basket.all_lines():
            weight += self.weigh_product(line.product) * line.quantity
        return weight



# pylint: disable=attribute-defined-outside-init
from django import http
from django.contrib import messages
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from django.views import generic

from oscar.core.loading import get_class, get_model

Product = get_model("catalogue", "Product")
ProductAlert = get_model("customer", "ProductAlert")
PageTitleMixin = get_class("customer.mixins", "PageTitleMixin")
ProductAlertForm = get_class("customer.forms", "ProductAlertForm")
AlertsDispatcher = get_class("customer.alerts.utils", "AlertsDispatcher")


class ProductAlertListView(PageTitleMixin, generic.ListView):
    model = ProductAlert
    template_name = "oscar/customer/alerts/alert_list.html"
    context_object_name = "alerts"
    page_title = _("Product Alerts")
    active_tab = "alerts"

    def get_queryset(self):
        return ProductAlert.objects.select_related().filter(
            user=self.request.user,
            date_closed=None,
        )


class ProductAlertCreateView(generic.CreateView):
    """
    View to create a new product alert based on a registered user
    or an email address provided by an anonymous user.
    """

    model = ProductAlert
    form_class = ProductAlertForm
    template_name = "oscar/customer/alerts/form.html"

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx["product"] = self.product
        ctx["alert_form"] = ctx.pop("form")
        return ctx

    def get(self, request, *args, **kwargs):
        product = get_object_or_404(Product, pk=self.kwargs["pk"])
        return http.HttpResponseRedirect(product.get_absolute_url())

    def post(self, request, *args, **kwargs):
        self.product = get_object_or_404(Product, pk=self.kwargs["pk"])
        return super().post(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        kwargs["product"] = self.product
        return kwargs

    def form_valid(self, form):
        response = super().form_valid(form)
        if self.object.is_anonymous:
            AlertsDispatcher().send_product_alert_confirmation_email_for_user(
                self.object
            )
        return response

    def get_success_url(self):
        if self.object.user:
            msg = _("An alert has been created")
        else:
            msg = _("A confirmation email has been sent to %s") % self.object.email
        messages.success(self.request, msg)
        return self.object.product.get_absolute_url()


class ProductAlertConfirmView(generic.RedirectView):
    permanent = False

    def get(self, request, *args, **kwargs):
        self.alert = get_object_or_404(ProductAlert, key=kwargs["key"])
        self.update_alert()
        return super().get(request, *args, **kwargs)

    def update_alert(self):
        if self.alert.can_be_confirmed:
            self.alert.confirm()
            messages.success(self.request, _("Your stock alert is now active"))
        else:
            messages.error(self.request, _("Your stock alert cannot be confirmed"))

    def get_redirect_url(self, *args, **kwargs):
        return self.alert.product.get_absolute_url()


class ProductAlertCancelView(generic.RedirectView):
    """
    This function allows canceling alerts by supplying the key (used for
    anonymously created alerts) or the pk (used for alerts created by a
    authenticated user).

    Specifying the redirect url is possible by supplying a 'next' GET
    parameter.  It defaults to showing the associated product page.
    """

    permanent = False

    def get(self, request, *args, **kwargs):
        if "key" in kwargs:
            self.alert = get_object_or_404(ProductAlert, key=kwargs["key"])
        elif "pk" in kwargs and request.user.is_authenticated:
            self.alert = get_object_or_404(
                ProductAlert, user=self.request.user, pk=kwargs["pk"]
            )
        else:
            raise Http404
        self.update_alert()
        return super().get(request, *args, **kwargs)

    def update_alert(self):
        if self.alert.can_be_cancelled:
            self.alert.cancel()
            messages.success(self.request, _("Your stock alert has been cancelled"))
        else:
            messages.error(self.request, _("Your stock alert cannot be cancelled"))

    def get_redirect_url(self, *args, **kwargs):
        return self.request.GET.get("next", self.alert.product.get_absolute_url())



from django.conf import settings
from django.contrib import messages
from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import gettext_lazy as _
from django.views.generic import CreateView, DetailView, ListView, View

from oscar.apps.catalogue.reviews.signals import review_added
from oscar.core.loading import get_classes, get_model
from oscar.core.utils import redirect_to_referrer

ProductReviewForm, VoteForm, SortReviewsForm = get_classes(
    "catalogue.reviews.forms", ["ProductReviewForm", "VoteForm", "SortReviewsForm"]
)

Vote = get_model("reviews", "vote")
ProductReview = get_model("reviews", "ProductReview")
Product = get_model("catalogue", "product")


class CreateProductReview(CreateView):
    template_name = "oscar/catalogue/reviews/review_form.html"
    model = ProductReview
    product_model = Product
    form_class = ProductReviewForm
    view_signal = review_added

    def dispatch(self, request, *args, **kwargs):
        # pylint: disable=attribute-defined-outside-init
        self.product = get_object_or_404(
            self.product_model, pk=kwargs["product_pk"], is_public=True
        )
        # check permission to leave review
        if not self.product.is_review_permitted(request.user):
            if self.product.has_review_by(request.user):
                message = _("You have already reviewed this product!")
            else:
                message = _("You can't leave a review for this product.")
            messages.warning(self.request, message)
            return redirect(self.product.get_absolute_url())

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["product"] = self.product
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["product"] = self.product
        kwargs["user"] = self.request.user
        return kwargs

    def form_valid(self, form):
        response = super().form_valid(form)
        self.send_signal(self.request, response, self.object)
        return response

    def get_success_url(self):
        messages.success(self.request, _("Thank you for reviewing this product"))
        return self.product.get_absolute_url()

    def send_signal(self, request, response, review):
        self.view_signal.send(
            sender=self,
            review=review,
            user=request.user,
            request=request,
            response=response,
        )


class ProductReviewDetail(DetailView):
    template_name = "oscar/catalogue/reviews/review_detail.html"
    context_object_name = "review"
    model = ProductReview

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["product"] = get_object_or_404(
            Product, pk=self.kwargs["product_pk"], is_public=True
        )
        return context


class AddVoteView(View):
    """
    Simple view for voting on a review.

    We use the URL path to determine the product and review and use a 'delta'
    POST variable to indicate it the vote is up or down.
    """

    def post(self, request, *args, **kwargs):
        product = get_object_or_404(
            Product, pk=self.kwargs["product_pk"], is_public=True
        )
        review = get_object_or_404(ProductReview, pk=self.kwargs["pk"])

        form = VoteForm(review, request.user, request.POST)
        if form.is_valid():
            if form.is_up_vote:
                review.vote_up(request.user)
            elif form.is_down_vote:
                review.vote_down(request.user)
            messages.success(request, _("Thanks for voting!"))
        else:
            for error_list in form.errors.values():
                for msg in error_list:
                    messages.error(request, msg)
        return redirect_to_referrer(request, product.get_absolute_url())


class ProductReviewList(ListView):
    """
    Browse reviews for a product
    """

    template_name = "oscar/catalogue/reviews/review_list.html"
    context_object_name = "reviews"
    model = ProductReview
    product_model = Product
    paginate_by = settings.OSCAR_REVIEWS_PER_PAGE

    def get_queryset(self):
        qs = self.model.objects.approved().filter(product=self.kwargs["product_pk"])
        # pylint: disable=attribute-defined-outside-init
        self.form = SortReviewsForm(self.request.GET)
        if self.request.GET and self.form.is_valid():
            sort_by = self.form.cleaned_data["sort_by"]
            if sort_by == SortReviewsForm.SORT_BY_RECENCY:
                return qs.order_by("-date_created")
        return qs.order_by("-score")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["product"] = get_object_or_404(
            self.product_model, pk=self.kwargs["product_pk"], is_public=True
        )
        context["form"] = self.form
        return context



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('partner', '0003_auto_20150604_1450'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='partner',
            options={'ordering': ('name', 'code'), 'verbose_name': 'Fulfillment partner', 'verbose_name_plural': 'Fulfillment partners', 'permissions': (('dashboard_access', 'Can access dashboard'),)},
        ),
    ]



from django import template

from oscar.core.utils import format_timedelta

register = template.Library()


@register.filter
def timedelta(td):
    """
    Return formatted timedelta value
    """
    return format_timedelta(td)



from decimal import ROUND_HALF_UP
from decimal import Decimal as D


def tax_subtract(price, tax_percentage=21):
    """Subtract the given `tax_percentage` from the given `price`."""
    if price is None:
        return None

    result = price / ((100 + tax_percentage) / D(100))
    return result.quantize(D("0.01"), ROUND_HALF_UP)


def tax_add(price, tax_percentage=21):
    """Add the given `tax_percentage` to the given `price`."""
    if price is None:
        return None

    result = price * ((100 + tax_percentage) / D(100))
    return result.quantize(D("0.01"), ROUND_HALF_UP)



import logging

from django.db import IntegrityError
from django.db.models import F
from django.dispatch import receiver

from oscar.apps.basket.signals import basket_addition
from oscar.apps.catalogue.signals import product_viewed
from oscar.apps.order.signals import order_placed
from oscar.apps.search.signals import user_search
from oscar.core.loading import get_model

ProductRecord = get_model("analytics", "ProductRecord")
UserProductView = get_model("analytics", "UserProductView")
UserRecord = get_model("analytics", "UserRecord")
UserSearch = get_model("analytics", "UserSearch")

# Helpers

logger = logging.getLogger("oscar.analytics")


def _update_counter(model, field_name, filter_kwargs, increment=1):
    """
    Efficiently updates a counter field by a given increment. Uses Django's
    update() call to fetch and update in one query.

    :param model: The model class of the recording model
    :param field_name: The name of the field to update
    :param filter_kwargs: Parameters to the ORM's filter() function to get the
                          correct instance
    """
    record = model.objects.filter(**filter_kwargs)
    affected = record.update(**{field_name: F(field_name) + increment})
    if not affected:
        filter_kwargs[field_name] = increment
        try:
            model.objects.create(**filter_kwargs)
        except IntegrityError:  # pragma: no cover
            # can happen in rare race condition
            logger.warning(
                "IntegrityError when updating analytics counter for %s", model
            )
            record.update(**{field_name: F(field_name) + increment})


def _record_products_in_order(order):
    for line in order.lines.all():
        _update_counter(
            ProductRecord,
            "num_purchases",
            {"product_id": line.product_id},
            line.quantity,
        )


def _record_user_order(user, order):
    try:
        record = UserRecord.objects.filter(user=user)
        affected = record.update(
            num_orders=F("num_orders") + 1,
            num_order_lines=F("num_order_lines") + order.num_lines,
            num_order_items=F("num_order_items") + order.num_items,
            total_spent=F("total_spent") + order.total_incl_tax,
            date_last_order=order.date_placed,
        )
        if not affected:
            UserRecord.objects.create(
                user=user,
                num_orders=1,
                num_order_lines=order.num_lines,
                num_order_items=order.num_items,
                total_spent=order.total_incl_tax,
                date_last_order=order.date_placed,
            )
    except IntegrityError:  # pragma: no cover
        logger.error("IntegrityError in analytics when recording a user order.")


# Receivers


# pylint: disable=unused-argument
@receiver(product_viewed)
def receive_product_view(sender, product, user, **kwargs):
    if kwargs.get("raw", False):
        return
    _update_counter(ProductRecord, "num_views", {"product": product})
    if user and user.is_authenticated:
        _update_counter(UserRecord, "num_product_views", {"user": user})
        UserProductView.objects.create(product=product, user=user)


# pylint: disable=unused-argument
@receiver(user_search)
def receive_product_search(sender, query, user, **kwargs):
    if user and user.is_authenticated and not kwargs.get("raw", False):
        UserSearch._default_manager.create(user=user, query=query)


# pylint: disable=unused-argument
@receiver(basket_addition)
def receive_basket_addition(sender, product, user, **kwargs):
    if kwargs.get("raw", False):
        return
    _update_counter(ProductRecord, "num_basket_additions", {"product": product})
    if user and user.is_authenticated:
        _update_counter(UserRecord, "num_basket_additions", {"user": user})


# pylint: disable=unused-argument
@receiver(order_placed)
def receive_order_placed(sender, order, user, **kwargs):
    if kwargs.get("raw", False):
        return
    _record_products_in_order(order)
    if user and user.is_authenticated:
        _record_user_order(user, order)



from oscar.core.loading import is_model_registered

from . import abstract_models

__all__ = []


if not is_model_registered("payment", "Transaction"):

    class Transaction(abstract_models.AbstractTransaction):
        pass

    __all__.append("Transaction")


if not is_model_registered("payment", "Source"):

    class Source(abstract_models.AbstractSource):
        pass

    __all__.append("Source")


if not is_model_registered("payment", "SourceType"):

    class SourceType(abstract_models.AbstractSourceType):
        pass

    __all__.append("SourceType")


if not is_model_registered("payment", "Bankcard"):

    class Bankcard(abstract_models.AbstractBankcard):
        pass

    __all__.append("Bankcard")



# Generated by Django 5.2 on 2025-10-01 14:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("order", "0019_order_analytics_tracked"),
    ]

    operations = [
        migrations.AlterField(
            model_name="order",
            name="analytics_tracked",
            field=models.BooleanField(default=False),
        ),
    ]



from oscar.apps.shipping import abstract_models
from oscar.core.loading import is_model_registered

__all__ = []


if not is_model_registered("shipping", "OrderAndItemCharges"):

    class OrderAndItemCharges(abstract_models.AbstractOrderAndItemCharges):
        pass

    __all__.append("OrderAndItemCharges")


if not is_model_registered("shipping", "WeightBased"):

    class WeightBased(abstract_models.AbstractWeightBased):
        pass

    __all__.append("WeightBased")


if not is_model_registered("shipping", "WeightBand"):

    class WeightBand(abstract_models.AbstractWeightBand):
        pass

    __all__.append("WeightBand")



import logging
import os
import shutil
from os.path import dirname, exists, join

from django.apps import apps

from oscar.core.application import OscarConfig


def create_local_app_folder(local_app_path):
    if exists(local_app_path):
        raise ValueError("There is already a '%s' folder! Aborting!" % local_app_path)
    for folder in subfolders(local_app_path):
        if not exists(folder):
            os.mkdir(folder)
            init_path = join(folder, "__init__.py")
            if not exists(init_path):
                create_file(init_path)


def subfolders(path):
    """
    Decompose a path string into a list of subfolders

    Eg Convert 'apps/dashboard/ranges' into
       ['apps', 'apps/dashboard', 'apps/dashboard/ranges']
    """
    folders = []
    while path not in ("/", ""):
        folders.append(path)
        path = dirname(path)
    folders.reverse()
    return folders


def inherit_app_config(local_app_folder_path, local_app_name, app_config):
    create_file(
        join(local_app_folder_path, "__init__.py"),
        "default_app_config = '{app_name}.apps.{app_config_class_name}'\n".format(
            app_name=local_app_name, app_config_class_name=app_config.__class__.__name__
        ),
    )
    create_file(
        join(local_app_folder_path, "apps.py"),
        "import {app_config_class_module} as apps\n\n\n"
        "class {app_config_class_name}(apps.{app_config_class_name}):\n"
        "    name = '{app_name}'\n".format(
            app_config_class_module=app_config.__module__,
            app_config_class_name=app_config.__class__.__name__,
            app_name=local_app_name,
        ),
    )


def fork_app(label, local_folder_path, local_app_subpackage=None, logger=None):
    """
    Create a custom version of one of Oscar's apps.

    The first argument is the app label of the Oscar app to fork.

    The second argument is a folder path, for where to copy the forked app.

    The third optional argument is the subpackage (inside the local folder path
    package) for the new app.
    """
    if logger is None:
        logger = logging.getLogger(__name__)

    # Check label is valid
    try:
        app_config = apps.get_app_config(label)
    except LookupError:
        raise ValueError("There is no app with the label '{}'".format(label))
    else:
        if not isinstance(app_config, OscarConfig):
            raise ValueError("There is no Oscar app with the label '{}'".format(label))

    # Remove trailing slash from folder path
    local_folder_path = local_folder_path.rstrip("/")

    # Check if local_folder_path is current folder
    if local_folder_path == ".":
        local_folder_path = ""

    local_apps_package = local_folder_path.lstrip("/").replace("/", ".")
    if local_app_subpackage is None:
        local_app_subpackage = app_config.name.replace("oscar.apps.", "")
        # In case this is a fork of a fork
        local_app_subpackage = local_app_subpackage.replace(local_apps_package, "")

    # Create folder
    local_app_subfolder_path = local_app_subpackage.replace(
        ".", "/"
    )  # eg 'dashboard/ranges'
    local_app_folder_path = join(local_folder_path, local_app_subfolder_path)
    logger.info("Creating package %s", local_app_folder_path)
    create_local_app_folder(local_app_folder_path)

    # Create minimum app files
    app_folder_path = app_config.path

    if exists(join(app_folder_path, "admin.py")):
        logger.info("Creating admin.py")
        create_file(
            join(local_app_folder_path, "admin.py"),
            "from {app_name}.admin import *  \n".format(app_name=app_config.name),
        )

    logger.info("Creating app config")
    local_app_name = (
        local_apps_package + ("." if local_apps_package else "") + local_app_subpackage
    )
    inherit_app_config(local_app_folder_path, local_app_name, app_config)

    # Only create models.py and migrations if they exist in the Oscar app
    models_file_path = join(app_folder_path, "models.py")
    if exists(models_file_path):
        logger.info("Creating models.py")
        create_file(
            join(local_app_folder_path, "models.py"),
            "from {app_name}.models import * \n".format(app_name=app_config.name),
        )

        migrations_subfolder_path = "migrations"
        migrations_folder_path = join(app_folder_path, migrations_subfolder_path)
        if exists(migrations_folder_path):
            logger.info("Creating %s folder", migrations_subfolder_path)
            local_migrations_folder_path = join(
                local_app_folder_path, migrations_subfolder_path
            )
            shutil.copytree(migrations_folder_path, local_migrations_folder_path)

    # Final step needs to be done by hand
    app_config_class_path = "{class_module}.{class_name}".format(
        class_module=app_config.__module__, class_name=app_config.__class__.__name__
    )
    local_app_config_class_path = "{local_app_name}.apps.{class_name}".format(
        local_app_name=local_app_name, class_name=app_config.__class__.__name__
    )
    msg = "Replace the entry '{}' with '{}' in INSTALLED_APPS".format(
        app_config_class_path, local_app_config_class_path
    )
    logger.info(msg)


def create_file(filepath, content=""):
    with open(filepath, "w", encoding="utf-8") as f:
        f.write(content)



import factory

from oscar.core.compat import get_user_model
from oscar.core.loading import get_model

__all__ = ["ProductAlertFactory", "UserFactory"]


class ProductAlertFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = get_model("customer", "ProductAlert")

    product = factory.SubFactory("oscar.test.factories.ProductFactory")
    user = factory.SubFactory("oscar.test.factories.customer.UserFactory")
    status = Meta.model.ACTIVE


class UserFactory(factory.django.DjangoModelFactory):
    username = factory.Sequence(lambda n: "the_j_meister nummer %d" % n)
    email = factory.Sequence(lambda n: "example_%s@example.com" % n)
    first_name = "joseph"
    last_name = "winterbottom"
    password = factory.PostGenerationMethodCall("set_password", "skelebrain")
    is_active = True
    is_superuser = False
    is_staff = False

    class Meta:
        model = get_user_model()



from django.urls import include, path

from .views import BandCreateView, BandDeleteView, BandListView, BandUpdateView

dashboard_urlpatterns = [
    path("band/create/", BandCreateView.as_view(), name="oscar-band-create"),
    path("band/", BandListView.as_view(), name="oscar-band-list"),
    # The RelatedFieldWidgetWrapper code does something funny with placeholder
    # urls, so it does need to match more than just a pk
    path("band/<str:pk>/update/", BandUpdateView.as_view(), name="oscar-band-update"),
    # The RelatedFieldWidgetWrapper code does something funny with placeholder
    # urls, so it does need to match more than just a pk
    path("band/<str:pk>/delete/", BandDeleteView.as_view(), name="oscar-band-delete"),
]

urlpatterns = [
    path(
        "dashboard/",
        include((dashboard_urlpatterns, "dashboard"), namespace="dashboard"),
    ),
]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import oscar.models.fields.autoslugfield
from django.utils.module_loading import import_string
from django.conf import settings
from decimal import Decimal

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)



class Migration(migrations.Migration):

    dependencies = [
        ('order', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Bankcard',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('card_type', models.CharField(max_length=128, verbose_name='Card Type')),
                ('name', models.CharField(max_length=255, verbose_name='Name', blank=True)),
                ('number', models.CharField(max_length=32, verbose_name='Number')),
                ('expiry_date', models.DateField(verbose_name='Expiry Date')),
                ('partner_reference', models.CharField(max_length=255, verbose_name='Partner Reference', blank=True)),
                ('user', models.ForeignKey(verbose_name='User', related_name='bankcards', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Bankcards',
                'verbose_name': 'Bankcard',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Source',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('currency', models.CharField(default='GBP', max_length=12, verbose_name='Currency')),
                ('amount_allocated', models.DecimalField(default=Decimal('0.00'), max_digits=12, decimal_places=2, verbose_name='Amount Allocated')),
                ('amount_debited', models.DecimalField(default=Decimal('0.00'), max_digits=12, decimal_places=2, verbose_name='Amount Debited')),
                ('amount_refunded', models.DecimalField(default=Decimal('0.00'), max_digits=12, decimal_places=2, verbose_name='Amount Refunded')),
                ('reference', models.CharField(max_length=128, verbose_name='Reference', blank=True)),
                ('label', models.CharField(max_length=128, verbose_name='Label', blank=True)),
                ('order', models.ForeignKey(verbose_name='Order', related_name='sources', to='order.Order', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Sources',
                'verbose_name': 'Source',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='SourceType',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, verbose_name='Name')),
                ('code', oscar.models.fields.autoslugfield.AutoSlugField(populate_from='name', unique=True, verbose_name='Code', editable=False, max_length=128, help_text='This is used within forms to identify this source type', blank=True)),
            ],
            options={
                'verbose_name_plural': 'Source Types',
                'verbose_name': 'Source Type',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('txn_type', models.CharField(max_length=128, verbose_name='Type', blank=True)),
                ('amount', models.DecimalField(max_digits=12, decimal_places=2, verbose_name='Amount')),
                ('reference', models.CharField(max_length=128, verbose_name='Reference', blank=True)),
                ('status', models.CharField(max_length=128, verbose_name='Status', blank=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('source', models.ForeignKey(verbose_name='Source', related_name='transactions', to='payment.Source', on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ['-date_created'],
                'verbose_name_plural': 'Transactions',
                'verbose_name': 'Transaction',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='source',
            name='source_type',
            field=models.ForeignKey(verbose_name='Source Type', related_name='sources', to='payment.SourceType', on_delete=models.CASCADE),
            preserve_default=True,
        ),
    ]



import json
from datetime import timedelta
from decimal import ROUND_UP
from decimal import Decimal as D

from django.contrib import messages
from django.contrib.auth import views as auth_views
from django.contrib.auth.forms import AuthenticationForm
from django.db.models import Avg, Count, Sum
from django.template.response import TemplateResponse
from django.urls import reverse_lazy
from django.utils.timezone import now
from django.views.generic import TemplateView

from oscar.core.compat import get_user_model
from oscar.core.loading import get_class, get_model

RelatedFieldWidgetWrapper = get_class("dashboard.widgets", "RelatedFieldWidgetWrapper")
ConditionalOffer = get_model("offer", "ConditionalOffer")
Voucher = get_model("voucher", "Voucher")
Basket = get_model("basket", "Basket")
StockAlert = get_model("partner", "StockAlert")
Product = get_model("catalogue", "Product")
Order = get_model("order", "Order")
Line = get_model("order", "Line")
User = get_user_model()


class IndexView(TemplateView):
    """
    An overview view which displays several reports about the shop.

    Supports the permission-based dashboard. It is recommended to add a
    :file:`oscar/dashboard/index_nonstaff.html` template because Oscar's
    default template will display potentially sensitive store information.
    """

    def get_template_names(self):
        if self.request.user.is_staff:
            return [
                "oscar/dashboard/index.html",
            ]
        else:
            return ["oscar/dashboard/index_nonstaff.html", "oscar/dashboard/index.html"]

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx.update(self.get_stats())
        return ctx

    def get_active_vouchers(self):
        """
        Get all active vouchers. The returned ``Queryset`` of vouchers
        is filtered by end date greater then the current date.
        """
        return Voucher.objects.filter(end_datetime__gt=now())

    def get_hourly_report(self, orders, hours=24, segments=10):
        """
        Get report of order revenue split up in hourly chunks. A report is
        generated for the last *hours* (default=24) from the current time.
        The report provides ``max_revenue`` of the hourly order revenue sum,
        ``y-range`` as the labelling for the y-axis in a template and
        ``order_total_hourly``, a list of properties for hourly chunks.
        *segments* defines the number of labelling segments used for the y-axis
        when generating the y-axis labels (default=10).
        """
        # Get datetime for 24 hours ago
        time_now = now().replace(minute=0, second=0)
        start_time = time_now - timedelta(hours=hours - 1)

        order_total_hourly = []
        for _ in range(0, hours, 2):
            end_time = start_time + timedelta(hours=2)
            hourly_orders = orders.filter(
                date_placed__gte=start_time, date_placed__lt=end_time
            )
            total = hourly_orders.aggregate(Sum("total_incl_tax"))[
                "total_incl_tax__sum"
            ] or D("0.0")
            order_total_hourly.append({"end_time": end_time, "total_incl_tax": total})
            start_time = end_time

        max_value = max([x["total_incl_tax"] for x in order_total_hourly])
        divisor = 1
        while divisor < max_value / 50:
            divisor *= 10
        max_value = (max_value / divisor).quantize(D("1"), rounding=ROUND_UP)
        max_value *= divisor
        if max_value:
            segment_size = (max_value) / D("100.0")
            for item in order_total_hourly:
                item["percentage"] = int(item["total_incl_tax"] / segment_size)

            y_range = []
            y_axis_steps = max_value / D(str(segments))
            for idx in reversed(range(segments + 1)):
                y_range.append(idx * y_axis_steps)
        else:
            y_range = []
            for item in order_total_hourly:
                item["percentage"] = 0

        ctx = {
            "order_total_hourly": order_total_hourly,
            "max_revenue": max_value,
            "y_range": y_range,
        }
        return ctx

    def get_stats(self):
        datetime_24hrs_ago = now() - timedelta(hours=24)

        orders = Order.objects.all()
        alerts = StockAlert.objects.all()
        baskets = Basket.objects.filter(status=Basket.OPEN)
        customers = User.objects.filter(orders__isnull=False).distinct()
        lines = Line.objects.filter()
        products = Product.objects.all()

        user = self.request.user
        if not user.is_staff:
            partners_ids = tuple(user.partners.values_list("id", flat=True))
            orders = orders.filter(lines__partner_id__in=partners_ids).distinct()
            alerts = alerts.filter(stockrecord__partner_id__in=partners_ids)
            baskets = baskets.filter(
                lines__stockrecord__partner_id__in=partners_ids
            ).distinct()
            customers = customers.filter(
                orders__lines__partner_id__in=partners_ids
            ).distinct()
            lines = lines.filter(partner_id__in=partners_ids)
            products = products.filter(stockrecords__partner_id__in=partners_ids)

        orders_last_day = orders.filter(date_placed__gt=datetime_24hrs_ago)

        open_alerts = alerts.filter(status=StockAlert.OPEN)
        closed_alerts = alerts.filter(status=StockAlert.CLOSED)

        total_lines_last_day = lines.filter(order__in=orders_last_day).count()
        stats = {
            "total_orders_last_day": orders_last_day.count(),
            "total_lines_last_day": total_lines_last_day,
            "average_order_costs": orders_last_day.aggregate(Avg("total_incl_tax"))[
                "total_incl_tax__avg"
            ]
            or D("0.00"),
            "total_revenue_last_day": orders_last_day.aggregate(Sum("total_incl_tax"))[
                "total_incl_tax__sum"
            ]
            or D("0.00"),
            "hourly_report_dict": self.get_hourly_report(orders),
            "total_customers_last_day": customers.filter(
                date_joined__gt=datetime_24hrs_ago,
            ).count(),
            "total_open_baskets_last_day": baskets.filter(
                date_created__gt=datetime_24hrs_ago
            ).count(),
            "total_products": products.count(),
            "total_open_stock_alerts": open_alerts.count(),
            "total_closed_stock_alerts": closed_alerts.count(),
            "total_customers": customers.count(),
            "total_open_baskets": baskets.count(),
            "total_orders": orders.count(),
            "total_lines": lines.count(),
            "total_revenue": orders.aggregate(Sum("total_incl_tax"))[
                "total_incl_tax__sum"
            ]
            or D("0.00"),
            "order_status_breakdown": orders.order_by("status")
            .values("status")
            .annotate(freq=Count("id")),
        }
        if user.is_staff:
            stats.update(
                offer_maps=(
                    ConditionalOffer.objects.filter(end_datetime__gt=now())
                    .values("offer_type")
                    .annotate(count=Count("id"))
                    .order_by("offer_type")
                ),
                total_vouchers=self.get_active_vouchers().count(),
            )
        return stats


class PopUpWindowMixin:
    @property
    def is_popup(self):
        return self.request.GET.get(
            RelatedFieldWidgetWrapper.IS_POPUP_VAR,
            self.request.POST.get(RelatedFieldWidgetWrapper.IS_POPUP_VAR),
        )

    @property
    def is_popup_var(self):
        return RelatedFieldWidgetWrapper.IS_POPUP_VAR

    def add_success_message(self, message):
        if not self.is_popup:
            messages.info(self.request, message)


class PopUpWindowCreateUpdateMixin(PopUpWindowMixin):
    @property
    def to_field(self):
        return self.request.GET.get(
            RelatedFieldWidgetWrapper.TO_FIELD_VAR,
            self.request.POST.get(RelatedFieldWidgetWrapper.TO_FIELD_VAR),
        )

    @property
    def to_field_var(self):
        return RelatedFieldWidgetWrapper.TO_FIELD_VAR

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)

        if self.is_popup:
            ctx["to_field"] = self.to_field
            ctx["to_field_var"] = self.to_field_var
            ctx["is_popup"] = self.is_popup
            ctx["is_popup_var"] = self.is_popup_var

        return ctx


class PopUpWindowCreateMixin(PopUpWindowCreateUpdateMixin):
    def popup_response(self, obj):
        if self.to_field:
            attr = str(self.to_field)
        else:
            attr = obj._meta.pk.attname
        value = obj.serializable_value(attr)
        popup_response_data = json.dumps(
            {
                "value": str(value),
                "obj": str(obj),
            }
        )
        return TemplateResponse(
            self.request,
            "oscar/dashboard/widgets/popup_response.html",
            {
                "popup_response_data": popup_response_data,
            },
        )


class PopUpWindowUpdateMixin(PopUpWindowCreateUpdateMixin):
    def popup_response(self, obj):
        opts = obj._meta
        if self.to_field:
            attr = str(self.to_field)
        else:
            attr = opts.pk.attname
        # Retrieve the `object_id` from the resolved pattern arguments.
        value = self.request.resolver_match.kwargs["pk"]
        new_value = obj.serializable_value(attr)
        popup_response_data = json.dumps(
            {
                "action": "change",
                "value": str(value),
                "obj": str(obj),
                "new_value": str(new_value),
            }
        )
        return TemplateResponse(
            self.request,
            "oscar/dashboard/widgets/popup_response.html",
            {
                "popup_response_data": popup_response_data,
            },
        )


class PopUpWindowDeleteMixin(PopUpWindowMixin):
    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)

        if self.is_popup:
            ctx["is_popup"] = self.is_popup
            ctx["is_popup_var"] = self.is_popup_var

        return ctx

    def delete(self, request, *args, **kwargs):
        """
        Calls the delete() method on the fetched object and then
        redirects to the success URL, or closes the popup, it it is one.
        """
        obj = self.get_object()

        response = super().delete(request, *args, **kwargs)

        if self.is_popup:
            obj_id = obj.pk
            popup_response_data = json.dumps(
                {
                    "action": "delete",
                    "value": str(obj_id),
                }
            )
            return TemplateResponse(
                request,
                "oscar/dashboard/widgets/popup_response.html",
                {
                    "popup_response_data": popup_response_data,
                },
            )
        else:
            return response

    def post(self, request, *args, **kwargs):
        """
        Calls the delete() method on the fetched object and then
        redirects to the success URL, or closes the popup, it it is one.
        """
        return self.delete(request, *args, **kwargs)


class LoginView(auth_views.LoginView):
    template_name = "oscar/dashboard/login.html"
    authentication_form = AuthenticationForm
    login_redirect_url = reverse_lazy("dashboard:index")

    def get_success_url(self):
        url = self.get_redirect_url()
        return url or self.login_redirect_url



import re
from calendar import monthrange
from datetime import date

from django import forms
from django.core import validators
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class, get_model
from oscar.forms.mixins import PhoneNumberMixin

from . import bankcards

Country = get_model("address", "Country")
BillingAddress = get_model("order", "BillingAddress")
Bankcard = get_model("payment", "Bankcard")
AbstractAddressForm = get_class("address.forms", "AbstractAddressForm")

# List of card names for all the card types supported in payment.bankcards
VALID_CARDS = set([card_type[0] for card_type in bankcards.CARD_TYPES])


class BankcardNumberField(forms.CharField):
    def __init__(self, *args, **kwargs):
        _kwargs = {
            "max_length": 20,
            "widget": forms.TextInput(attrs={"autocomplete": "off"}),
            "label": _("Card number"),
        }
        if "types" in kwargs:
            self.accepted_cards = set(kwargs.pop("types"))
            difference = self.accepted_cards - VALID_CARDS
            if difference:
                raise ImproperlyConfigured(
                    "The following accepted_cards are unknown: %s" % difference
                )

        _kwargs.update(kwargs)
        super().__init__(*args, **_kwargs)

    def clean(self, value):
        """
        Check if given CC number is valid and one of the
        card types we accept
        """
        non_decimal = re.compile(r"\D+")
        value = non_decimal.sub("", (value or "").strip())

        if value and not bankcards.luhn(value):
            raise forms.ValidationError(_("Please enter a valid credit card number."))

        if hasattr(self, "accepted_cards"):
            card_type = bankcards.bankcard_type(value)
            if card_type not in self.accepted_cards:
                raise forms.ValidationError(_("%s cards are not accepted." % card_type))

        return super().clean(value)


class BankcardMonthWidget(forms.MultiWidget):
    """
    Widget containing two select boxes for selecting the month and year
    """

    def decompress(self, value):
        return [value.month, value.year] if value else [None, None]

    def format_output(self, rendered_widgets):
        html = " ".join(rendered_widgets)
        return '<span style="white-space: nowrap">%s</span>' % html


# pylint: disable=W0223
class BankcardMonthField(forms.MultiValueField):
    """
    A modified version of the snippet: http://djangosnippets.org/snippets/907/
    """

    default_error_messages = {
        "invalid_month": _("Enter a valid month."),
        "invalid_year": _("Enter a valid year."),
    }
    num_years = 5

    def __init__(self, *args, **kwargs):
        # Allow the number of years to be specified
        if "num_years" in kwargs:
            self.num_years = kwargs.pop("num_years")

        errors = self.default_error_messages.copy()
        if "error_messages" in kwargs:
            errors.update(kwargs["error_messages"])

        fields = (
            forms.ChoiceField(
                choices=self.month_choices(),
                error_messages={"invalid": errors["invalid_month"]},
            ),
            forms.ChoiceField(
                choices=self.year_choices(),
                error_messages={"invalid": errors["invalid_year"]},
            ),
        )
        if "widget" not in kwargs:
            kwargs["widget"] = BankcardMonthWidget(
                widgets=[fields[0].widget, fields[1].widget]
            )
        super().__init__(fields, *args, **kwargs)

    def month_choices(self):
        return []

    def year_choices(self):
        return []


class BankcardExpiryMonthField(BankcardMonthField):
    num_years = 10

    def __init__(self, *args, **kwargs):
        today = date.today()
        _kwargs = {
            "required": True,
            "label": _("Valid to"),
            "initial": ["%.2d" % today.month, today.year],
        }
        _kwargs.update(kwargs)
        super().__init__(*args, **_kwargs)

    def month_choices(self):
        return [("%.2d" % x, "%.2d" % x) for x in range(1, 13)]

    def year_choices(self):
        return [
            (x, x) for x in range(date.today().year, date.today().year + self.num_years)
        ]

    def clean(self, value):
        expiry_date = super().clean(value)
        if expiry_date and date.today() > expiry_date:
            raise forms.ValidationError(
                _("The expiration date you entered is in the past.")
            )
        return expiry_date

    def compress(self, data_list):
        if data_list:
            if data_list[1] in validators.EMPTY_VALUES:
                error = self.error_messages["invalid_year"]
                raise forms.ValidationError(error)
            if data_list[0] in validators.EMPTY_VALUES:
                error = self.error_messages["invalid_month"]
                raise forms.ValidationError(error)
            year = int(data_list[1])
            month = int(data_list[0])
            # find last day of the month
            day = monthrange(year, month)[1]
            return date(year, month, day)
        return None


class BankcardStartingMonthField(BankcardMonthField):
    def __init__(self, *args, **kwargs):
        _kwargs = {
            "required": False,
            "label": _("Valid from"),
        }
        _kwargs.update(kwargs)
        super().__init__(*args, **_kwargs)

    def month_choices(self):
        months = [("%.2d" % x, "%.2d" % x) for x in range(1, 13)]
        months.insert(0, ("", "--"))
        return months

    def year_choices(self):
        today = date.today()
        years = [(x, x) for x in range(today.year - self.num_years, today.year + 1)]
        years.insert(0, ("", "--"))
        return years

    def clean(self, value):
        starting_date = super().clean(value)
        if starting_date and date.today() < starting_date:
            raise forms.ValidationError(
                _("The starting date you entered is in the future.")
            )
        return starting_date

    def compress(self, data_list):
        if data_list:
            if data_list[1] in validators.EMPTY_VALUES:
                error = self.error_messages["invalid_year"]
                raise forms.ValidationError(error)
            if data_list[0] in validators.EMPTY_VALUES:
                error = self.error_messages["invalid_month"]
                raise forms.ValidationError(error)
            year = int(data_list[1])
            month = int(data_list[0])
            return date(year, month, 1)
        return None


class BankcardCCVField(forms.RegexField):
    def __init__(self, *args, **kwargs):
        _kwargs = {
            "required": True,
            "label": _("CCV number"),
            "widget": forms.TextInput(attrs={"size": "5"}),
            "error_messages": {"invalid": _("Please enter a 3 or 4 digit number")},
            "help_text": _(
                "This is the 3 or 4 digit security number "
                "on the back of your bankcard"
            ),
        }
        _kwargs.update(kwargs)
        super().__init__(r"^\d{3,4}$", *args, **_kwargs)

    def clean(self, value):
        if value is not None:
            value = value.strip()
        return super().clean(value)


class BankcardForm(forms.ModelForm):
    # By default, this number field will accept any number. The only validation
    # is whether it passes the luhn check. If you wish to only accept certain
    # types of card, you can pass a types kwarg to BankcardNumberField, e.g.
    #
    # BankcardNumberField(types=[bankcards.VISA, bankcards.VISA_ELECTRON,])

    number = BankcardNumberField()
    ccv = BankcardCCVField()
    start_month = BankcardStartingMonthField()
    expiry_month = BankcardExpiryMonthField()

    class Meta:
        model = Bankcard
        fields = ("number", "start_month", "expiry_month", "ccv")

    def clean(self):
        data = self.cleaned_data
        number, ccv = data.get("number"), data.get("ccv")
        if number and ccv:
            if bankcards.is_amex(number) and len(ccv) != 4:
                raise forms.ValidationError(
                    _("American Express cards use a 4 digit security code")
                )
        return data

    def save(self, *args, **kwargs):
        # It doesn't really make sense to save directly from the form as saving
        # will obfuscate some of the card details which you normally need to
        # pass to a payment gateway.  Better to use the bankcard property below
        # to get the cleaned up data, then once you've used the sensitive
        # details, you can save.
        raise RuntimeError("Don't save bankcards directly from form")

    @property
    def bankcard(self):
        """
        Return an instance of the Bankcard model (unsaved)
        """
        return Bankcard(
            number=self.cleaned_data["number"],
            expiry_date=self.cleaned_data["expiry_month"],
            start_date=self.cleaned_data["start_month"],
            ccv=self.cleaned_data["ccv"],
        )


class BillingAddressForm(PhoneNumberMixin, AbstractAddressForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_country_queryset()

    def set_country_queryset(self):
        self.fields["country"].queryset = Country._default_manager.all()

    class Meta:
        model = BillingAddress
        fields = [
            "first_name",
            "last_name",
            "line1",
            "line2",
            "line3",
            "line4",
            "state",
            "postcode",
            "country",
        ]



import logging

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.mail import EmailMessage, EmailMultiAlternatives

from oscar.core.loading import get_model

CommunicationEventType = get_model("communication", "CommunicationEventType")
Email = get_model("communication", "Email")
Notification = get_model("communication", "Notification")


class Dispatcher(object):
    def __init__(self, logger=None, mail_connection=None):
        if not logger:
            logger = logging.getLogger(__name__)
        self.logger = logger
        # Supply a mail_connection if you want the dispatcher to use that
        # instead of opening a new one.
        self.mail_connection = mail_connection

    # Public API methods

    def dispatch_direct_messages(self, recipient_email, messages, attachments=None):
        """
        Dispatch one-off messages to explicitly specified recipient email.
        """
        if messages["subject"] and (messages["body"] or messages["html"]):
            return self.send_email_messages(
                recipient_email, messages, attachments=attachments
            )

    def dispatch_anonymous_messages(self, email, messages, attachments=None):
        dispatched_messages = {}
        if email:
            dispatched_messages["email"] = (
                self.send_email_messages(email, messages, attachments=attachments),
                None,
            )
        return dispatched_messages

    def dispatch_user_messages(self, user, messages, attachments=None):
        """
        Send messages to a site user
        """
        dispatched_messages = {}
        if messages["subject"] and (messages["body"] or messages["html"]):
            dispatched_messages["email"] = self.send_user_email_messages(
                user, messages, attachments
            )
        if messages["sms"]:
            dispatched_messages["sms"] = self.send_text_message(user, messages["sms"])
        return dispatched_messages

    def notify_user(self, user, subject, **kwargs):
        """
        Send a simple notification to a user
        """
        Notification.objects.create(recipient=user, subject=subject, **kwargs)

    def notify_users(self, users, subject, **kwargs):
        """
        Send a simple notification to an iterable of users
        """
        for user in users:
            self.notify_user(user, subject, **kwargs)

    # Internal

    def create_email(self, user, messages, email):
        """
        Create ``Email`` instance in database for logging purposes.
        """
        if email and user.is_authenticated:
            return Email.objects.create(
                user=user,
                email=user.email,
                subject=email.subject,
                body_text=email.body,
                body_html=messages["html"],
            )

    def send_user_email_messages(self, user, messages, attachments=None):
        """
        Send message to the registered user / customer and collect data in database.
        """
        if not user.email:
            self.logger.warning(
                "Unable to send email messages as user #%d has no email address",
                user.id,
            )
            return None

        email = self.send_email_messages(user.email, messages, attachments=attachments)

        if settings.OSCAR_SAVE_SENT_EMAILS_TO_DB:
            self.create_email(user, messages, email)

        return email

    def send_email_messages(
        self, recipient_email, messages, from_email=None, attachments=None
    ):
        """
        Send email to recipient, HTML attachment optional.
        """
        from_email = from_email or settings.OSCAR_FROM_EMAIL

        content_attachments, file_attachments = self.prepare_attachments(attachments)

        # Determine whether we are sending a HTML version too
        if messages["html"]:
            email = EmailMultiAlternatives(
                messages["subject"],
                messages["body"],
                from_email=from_email,
                to=[recipient_email],
                attachments=content_attachments,
            )
            email.attach_alternative(messages["html"], "text/html")
        else:
            email = EmailMessage(
                messages["subject"],
                messages["body"],
                from_email=from_email,
                to=[recipient_email],
                attachments=content_attachments,
            )
        for attachment in file_attachments:
            email.attach_file(attachment)

        self.logger.info("Sending email to %s", recipient_email)

        if self.mail_connection:
            self.mail_connection.send_messages([email])
        else:
            email.send()

        return email

    def send_text_message(self, user, event_type):
        raise NotImplementedError

    def prepare_attachments(self, attachments):
        """
        Two types of attachments can be attached to emails:

            * "Content" attachment is one of:
                * instance of ``MIMEBase`` (from ``email.mime.base``);
                * list ``[filename, content, mimetype]``;

            * "File" attachment is a path to file from an instance of
              ``FileField`` based fields.

        "Content" and "file" attachments attached to emails differently.
        """
        content_attachments = []
        file_attachments = []
        if attachments is not None:
            for attachment in attachments:
                if isinstance(attachment, str):
                    file_attachments.append(attachment)
                else:
                    content_attachments.append(attachment)

        return content_attachments, file_attachments

    def get_base_context(self, **kwargs):
        """
        Return context that is common to all emails
        """
        request = kwargs.get("request")
        return dict(kwargs, site=Site.objects.get_current(request))

    def get_messages(self, event_code, extra_context=None):
        """
        Return rendered messages
        """
        if extra_context is None:
            extra_context = {}
        context = self.get_base_context(**extra_context)

        msgs = CommunicationEventType.objects.get_and_render(event_code, context)
        return msgs



from decimal import Decimal as D
from decimal import InvalidOperation

from babel.numbers import format_currency
from django import template
from django.conf import settings
from django.utils.translation import get_language, to_locale

register = template.Library()


@register.filter(name="currency")
def currency(value, currency_format=None):
    """
    Format decimal value as currency
    """
    if currency_format is None:
        currency_format = settings.OSCAR_DEFAULT_CURRENCY

    try:
        value = D(value)
    except (TypeError, InvalidOperation):
        return ""
    # Using Babel's currency formatting
    # http://babel.pocoo.org/en/latest/api/numbers.html#babel.numbers.format_currency
    OSCAR_CURRENCY_FORMAT = getattr(settings, "OSCAR_CURRENCY_FORMAT", None)
    kwargs = {
        "currency": currency_format,
        "locale": to_locale(get_language() or settings.LANGUAGE_CODE),
    }
    if isinstance(OSCAR_CURRENCY_FORMAT, dict):
        kwargs.update(OSCAR_CURRENCY_FORMAT.get(currency_format, {}))
    else:
        kwargs["format"] = OSCAR_CURRENCY_FORMAT
    return format_currency(value, **kwargs)



# pylint: disable=E5142, E5141
import csv

from django.conf import settings
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured

from oscar.core.loading import get_model

# A setting that can be used in foreign key declarations
AUTH_USER_MODEL = getattr(settings, "AUTH_USER_MODEL", "auth.User")
try:
    AUTH_USER_APP_LABEL, AUTH_USER_MODEL_NAME = AUTH_USER_MODEL.rsplit(".", 1)
except ValueError:
    raise ImproperlyConfigured(
        "AUTH_USER_MODEL must be of the form 'app_label.model_name'"
    )


def get_user_model():
    """
    Return the User model. Doesn't require the app cache to be fully
    initialised.

    This used to live in compat to support both Django 1.4's fixed User model
    and custom user models introduced thereafter.
    Support for Django 1.4 has since been dropped in Oscar, but our
    get_user_model remains because code relies on us annotating the _meta class
    with the additional fields, and other code might rely on it as well.
    """

    try:
        model = get_model(AUTH_USER_APP_LABEL, AUTH_USER_MODEL_NAME)
    except LookupError:
        # Convert exception to an ImproperlyConfigured exception for
        # backwards compatibility with previous Oscar versions and the
        # original get_user_model method in Django.
        raise ImproperlyConfigured(
            "AUTH_USER_MODEL refers to model '%s' that has not been installed"
            % settings.AUTH_USER_MODEL
        )

    # Test if user model has any custom fields and add attributes to the _meta
    # class
    core_fields = set([f.name for f in User._meta.fields])
    model_fields = set([f.name for f in model._meta.fields])
    new_fields = model_fields.difference(core_fields)
    model._meta.has_additional_fields = len(new_fields) > 0
    model._meta.additional_fields = new_fields

    return model


def existing_user_fields(fields):
    """
    Starting with Django 1.6, the User model can be overridden  and it is no
    longer safe to assume the User model has certain fields. This helper
    function assists in writing portable forms Meta.fields definitions
    when those contain fields on the User model

    Usage:
    class UserForm(forms.Form):
        ...
        class Meta:
            # won't break if first_name is not defined on User model
            fields = existing_user_fields(['first_name', 'last_name'])
    """
    user_fields = get_user_model()._meta.fields
    user_field_names = [field.name for field in user_fields]
    return [field for field in fields if field in user_field_names]


class UnicodeCSVWriter:
    """
    MS Excel compatible CSV writer. Supports two modes:
    * Writing to an open file or file-like object:
      writer = UnicodeCSVWriter(open_file=your_file)
      ...
      your_file.close()
    * Writing to a new file:
      with UnicodeCSVWriter(filename=filename) as writer:
          ...
    """

    def __init__(
        self, filename=None, open_file=None, dialect=csv.excel, encoding="utf-8", **kw
    ):
        if filename is open_file is None:
            raise ImproperlyConfigured(
                "You need to specify either a filename or an open file"
            )
        self.filename = filename
        self.f = open_file
        self.dialect = dialect
        self.encoding = encoding
        self.kw = kw
        self.writer = None

        if self.f:
            self.add_bom(self.f)

    def __enter__(self):
        assert self.filename is not None
        self.f = open(self.filename, "wt", encoding=self.encoding, newline="")
        self.add_bom(self.f)
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        assert self.filename is not None
        if self.filename is not None:
            self.f.close()

    # pylint: disable=unused-argument
    def add_bom(self, f):
        # If encoding is UTF-8, insert a Byte Order Mark at the start of the
        # file for compatibility with MS Excel.
        if self.encoding == "utf-8" and getattr(
            settings, "OSCAR_CSV_INCLUDE_BOM", False
        ):
            self.f.write("\ufeff")

    def writerow(self, row):
        if self.writer is None:
            self.writer = csv.writer(self.f, dialect=self.dialect, **self.kw)
        self.writer.writerow(list(row))

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)



# Generated by Django 3.0.8 on 2020-08-01 07:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('communication', '0003_remove_notification_category_make_code_uppercase'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='communicationeventtype',
            options={'ordering': ['name'], 'verbose_name': 'Communication event type', 'verbose_name_plural': 'Communication event types'},
        ),
        migrations.AlterModelOptions(
            name='email',
            options={'ordering': ['-date_sent'], 'verbose_name': 'Email', 'verbose_name_plural': 'Emails'},
        ),
        migrations.AlterField(
            model_name='communicationeventtype',
            name='name',
            field=models.CharField(db_index=True, max_length=255, verbose_name='Name'),
        ),
    ]



from django import shortcuts
from django.contrib import messages
from django.contrib.sessions.serializers import JSONSerializer
from django.core.exceptions import ObjectDoesNotExist
from django.http import JsonResponse, QueryDict
from django.shortcuts import redirect
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView, View
from extra_views import ModelFormSetView

from oscar.apps.basket.signals import basket_addition, voucher_addition, voucher_removal
from oscar.core import ajax
from oscar.core.loading import get_class, get_classes, get_model
from oscar.core.utils import is_ajax, redirect_to_referrer, safe_referrer

Applicator = get_class("offer.applicator", "Applicator")
(BasketLineForm, AddToBasketForm, BasketVoucherForm, SavedLineForm) = get_classes(
    "basket.forms",
    ("BasketLineForm", "AddToBasketForm", "BasketVoucherForm", "SavedLineForm"),
)
BasketLineFormSet, SavedLineFormSet = get_classes(
    "basket.formsets", ("BasketLineFormSet", "SavedLineFormSet")
)
Repository = get_class("shipping.repository", "Repository")

OrderTotalCalculator = get_class("checkout.calculators", "OrderTotalCalculator")
BasketMessageGenerator = get_class("basket.utils", "BasketMessageGenerator")
SurchargeApplicator = get_class("checkout.applicator", "SurchargeApplicator")


class BasketView(ModelFormSetView):
    model = get_model("basket", "Line")
    basket_model = get_model("basket", "Basket")
    formset_class = BasketLineFormSet
    form_class = BasketLineForm
    factory_kwargs = {"extra": 0, "can_delete": True}
    template_name = "oscar/basket/basket.html"

    def get_formset_kwargs(self):
        kwargs = super().get_formset_kwargs()
        kwargs["strategy"] = self.request.strategy
        return kwargs

    def get_queryset(self):
        """
        Return list of :py:class:`Line <oscar.apps.basket.abstract_models.AbstractLine>`
        instances associated with the current basket.
        """
        return self.request.basket.all_lines()

    # pylint: disable=unused-argument
    def get_shipping_methods(self, basket):
        return Repository().get_shipping_methods(
            basket=self.request.basket, user=self.request.user, request=self.request
        )

    def get_default_shipping_address(self):
        if self.request.user.is_authenticated:
            return self.request.user.addresses.filter(
                is_default_for_shipping=True
            ).first()

    # pylint: disable=unused-argument
    def get_default_shipping_method(self, basket):
        return Repository().get_default_shipping_method(
            basket=self.request.basket,
            user=self.request.user,
            request=self.request,
            shipping_addr=self.get_default_shipping_address(),
        )

    def get_basket_warnings(self, basket):
        """
        Return a list of warnings that apply to this basket
        """
        warnings = []
        for line in basket.all_lines():
            warning = line.get_warning()
            if warning:
                warnings.append(warning)
        return warnings

    def get_upsell_messages(self, basket):
        offers = Applicator().get_offers(basket, self.request.user, self.request)
        applied_offers = list(basket.offer_applications.offers.values())
        msgs = []
        for offer in offers:
            if (
                offer.is_condition_partially_satisfied(basket)
                and offer not in applied_offers
            ):
                data = {"message": offer.get_upsell_message(basket), "offer": offer}
                msgs.append(data)
        return msgs

    def get_basket_voucher_form(self):
        """
        This is a separate method so that it's easy to e.g. not return a form
        if there are no vouchers available.
        """
        return BasketVoucherForm()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["voucher_form"] = self.get_basket_voucher_form()

        # Shipping information is included to give an idea of the total order
        # cost.  It is also important for PayPal Express where the customer
        # gets redirected away from the basket page and needs to see what the
        # estimated order total is beforehand.
        context["shipping_methods"] = self.get_shipping_methods(self.request.basket)
        method = self.get_default_shipping_method(self.request.basket)
        context["shipping_method"] = method
        shipping_charge = method.calculate(self.request.basket)
        context["shipping_charge"] = shipping_charge
        if method.is_discounted:
            excl_discount = method.calculate_excl_discount(self.request.basket)
            context["shipping_charge_excl_discount"] = excl_discount
        context["basket_warnings"] = self.get_basket_warnings(self.request.basket)
        context["upsell_messages"] = self.get_upsell_messages(self.request.basket)

        if self.request.user.is_authenticated:
            try:
                saved_basket = self.basket_model.saved.get(owner=self.request.user)
            except self.basket_model.DoesNotExist:
                pass
            else:
                saved_basket.strategy = self.request.basket.strategy
                if not saved_basket.is_empty:
                    saved_queryset = saved_basket.all_lines()
                    formset = SavedLineFormSet(
                        strategy=self.request.strategy,
                        basket=self.request.basket,
                        queryset=saved_queryset,
                        prefix="saved",
                    )
                    context["saved_formset"] = formset

        surcharges = SurchargeApplicator(
            self.request, context
        ).get_applicable_surcharges(
            self.request.basket, shipping_charge=shipping_charge
        )
        context["surcharges"] = surcharges
        context["order_total"] = OrderTotalCalculator().calculate(
            self.request.basket, shipping_charge, surcharges=surcharges
        )
        return context

    def get_success_url(self):
        return safe_referrer(self.request, "basket:summary")

    def formset_valid(self, formset):
        # Store offers before any changes are made so we can inform the user of
        # any changes
        offers_before = self.request.basket.applied_offers()
        save_for_later = False

        # Keep a list of messages - we don't immediately call
        # django.contrib.messages as we may be returning an AJAX response in
        # which case we pass the messages back in a JSON payload.
        flash_messages = ajax.FlashMessages()

        for form in formset:
            if hasattr(form, "cleaned_data") and form.cleaned_data.get(
                "save_for_later", False
            ):
                line = form.instance
                if self.request.user.is_authenticated:
                    self.move_line_to_saved_basket(line)

                    msg = render_to_string(
                        "oscar/basket/messages/line_saved.html", {"line": line}
                    )
                    flash_messages.info(msg)

                    save_for_later = True
                else:
                    msg = _("You can't save an item for later if you're not logged in!")
                    flash_messages.error(msg)
                    return redirect(self.get_success_url())

        if save_for_later:
            # No need to call super if we're moving lines to the saved basket
            response = redirect(self.get_success_url())
        else:
            # Save changes to basket as per normal
            response = super().formset_valid(formset)

        # If AJAX submission, don't redirect but reload the basket content HTML
        if is_ajax(self.request):
            # Reload basket and apply offers again
            self.request.basket = get_model("basket", "Basket").objects.get(
                id=self.request.basket.id
            )
            self.request.basket.strategy = self.request.strategy
            Applicator().apply(self.request.basket, self.request.user, self.request)
            offers_after = self.request.basket.applied_offers()

            for level, msg in BasketMessageGenerator().get_messages(
                self.request.basket, offers_before, offers_after, include_buttons=False
            ):
                flash_messages.add_message(level, msg)

            # Reload formset - we have to remove the POST fields from the
            # kwargs as, if they are left in, the formset won't construct
            # correctly as there will be a state mismatch between the
            # management form and the database.
            kwargs = self.get_formset_kwargs()
            del kwargs["data"]
            del kwargs["files"]
            if "queryset" in kwargs:
                del kwargs["queryset"]
            formset = self.get_formset()(queryset=self.get_queryset(), **kwargs)
            ctx = self.get_context_data(formset=formset, basket=self.request.basket)
            return self.json_response(ctx, flash_messages)

        BasketMessageGenerator().apply_messages(self.request, offers_before)

        return response

    def json_response(self, ctx, flash_messages):
        basket_html = render_to_string(
            "oscar/basket/partials/basket_content.html",
            context=ctx,
            request=self.request,
        )

        return JsonResponse(
            {"content_html": basket_html, "messages": flash_messages.as_dict()}
        )

    def move_line_to_saved_basket(self, line):
        saved_basket, _ = get_model("basket", "basket").saved.get_or_create(
            owner=self.request.user
        )
        saved_basket.merge_line(line)

    def formset_invalid(self, formset):
        has_deletion = any(formset._should_delete_form(form) for form in formset.forms)
        has_no_invalid_non_deletion = all(
            form.is_valid() or formset._should_delete_form(form)
            for form in formset.forms
        )
        if has_deletion:
            self.remove_deleted_forms(formset)
            if has_no_invalid_non_deletion:
                return self.formset_valid(formset)

        flash_messages = ajax.FlashMessages()
        flash_messages.warning(
            _(
                "Your basket has got some issues. "
                "Please correct any validation errors below."
            )
        )

        if is_ajax(self.request):
            ctx = self.get_context_data(formset=formset, basket=self.request.basket)
            return self.json_response(ctx, flash_messages)

        flash_messages.apply_to_request(self.request)
        return super().formset_invalid(formset)

    def remove_deleted_forms(self, formset):
        """
        Removes forms marked for deletion, from the formset, as well as deletes
        their model instance objects; and modifies the formset's request data,
        to match the state of the data in the database, for the remaining forms.

        This is useful for redisplaying a formset containing other invalid
        forms, after deleting one of the forms from it.
        """
        form_data = {}
        form_index = 0
        for form in formset.forms:
            # Delete forms marked for deletion, and retain the request data
            # for the other forms
            if formset._should_delete_form(form):
                if form.instance.id is not None:
                    form.instance.delete()
            else:
                old_form_prefix = form.prefix
                new_form_prefix = formset.add_prefix(form_index)
                for field_name in form.fields:
                    form.prefix = old_form_prefix
                    old_prefixed_field_name = form.add_prefix(field_name)
                    form.prefix = new_form_prefix
                    new_prefixed_field_name = form.add_prefix(field_name)
                    try:
                        form_data[new_prefixed_field_name] = formset.data[
                            old_prefixed_field_name
                        ]
                    except KeyError:
                        pass
                form_index += 1
        for field_name in formset.management_form.fields:
            prefixed_field_name = formset.management_form.add_prefix(field_name)
            if field_name in ["INITIAL_FORMS", "TOTAL_FORMS"]:
                form_data[prefixed_field_name] = str(form_index)
            else:
                form_data[prefixed_field_name] = formset.data[prefixed_field_name]

        query_dict = QueryDict(mutable=True)
        query_dict.update(form_data)
        formset.data = query_dict
        # Clear cached values, so that they are recomputed using the modified
        # request data
        del formset.management_form
        del formset.forms
        # Clean the formset's modified request data
        formset.full_clean()


class BasketAddView(FormView):
    """
    Handles the add-to-basket submissions, which are triggered from various
    parts of the site. The add-to-basket form is loaded into templates using
    a templatetag from :py:mod:`oscar.templatetags.basket_tags`.
    """

    form_class = AddToBasketForm
    product_model = get_model("catalogue", "product")
    add_signal = basket_addition
    http_method_names = ["post"]

    def post(self, request, *args, **kwargs):
        # pylint: disable=W0201
        self.product = shortcuts.get_object_or_404(self.product_model, pk=kwargs["pk"])
        return super().post(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["basket"] = self.request.basket
        kwargs["product"] = self.product
        return kwargs

    def form_invalid(self, form):
        msgs = []
        for error in form.errors.values():
            msgs.append(error.as_text())
        clean_msgs = [m.replace("* ", "") for m in msgs if m.startswith("* ")]
        messages.error(self.request, ",".join(clean_msgs))

        # We serialize the POST data with JSONSerializer before adding it to the session.
        # Without this, we could expose the site to a security vulnerability
        # if the SESSION_SERIALIZER has been configured to 'django.contrib.sessions.serializers.PickleSerializer'.
        # see: https://docs.djangoproject.com/en/3.2/topics/http/sessions/#cookie-session-backend
        serialized_data = JSONSerializer().dumps(self.request.POST)
        self.request.session["add_to_basket_form_post_data_%s" % self.product.pk] = (
            serialized_data.decode("latin-1")
        )

        return redirect_to_referrer(self.request, "basket:summary")

    def form_valid(self, form):
        offers_before = self.request.basket.applied_offers()

        # pylint: disable=W0201
        self.line, self.line_created = self.request.basket.add_product(
            form.product, form.cleaned_data["quantity"], form.cleaned_options()
        )

        messages.success(
            self.request, self.get_success_message(form), extra_tags="safe noicon"
        )

        # Check for additional offer messages
        BasketMessageGenerator().apply_messages(self.request, offers_before)

        # Send signal for basket addition
        self.add_signal.send(
            sender=self,
            product=form.product,
            user=self.request.user,
            request=self.request,
        )

        return super().form_valid(form)

    def get_success_message(self, form):
        return render_to_string(
            "oscar/basket/messages/addition.html",
            {"product": form.product, "quantity": form.cleaned_data["quantity"]},
        )

    def get_success_url(self):
        post_url = self.request.POST.get("next")
        if post_url and url_has_allowed_host_and_scheme(
            post_url, self.request.get_host()
        ):
            return post_url
        return safe_referrer(self.request, "basket:summary")


class VoucherAddView(FormView):
    form_class = BasketVoucherForm
    voucher_model = get_model("voucher", "voucher")
    add_signal = voucher_addition

    def get(self, request, *args, **kwargs):
        return redirect("basket:summary")

    def apply_voucher_to_basket(self, voucher):
        if voucher.is_expired():
            messages.error(
                self.request,
                _("The '%(code)s' voucher has expired") % {"code": voucher.code},
            )
            return

        if not voucher.is_active():
            messages.error(
                self.request,
                _("The '%(code)s' voucher is not active") % {"code": voucher.code},
            )
            return

        is_available, message = voucher.is_available_to_user(self.request.user)
        if not is_available:
            messages.error(self.request, message)
            return

        self.request.basket.vouchers.add(voucher)

        # Raise signal
        self.add_signal.send(sender=self, basket=self.request.basket, voucher=voucher)

        # Recalculate discounts to see if the voucher gives any
        Applicator().apply(self.request.basket, self.request.user, self.request)
        discounts_after = self.request.basket.offer_applications

        # Look for discounts from this new voucher
        found_discount = False
        for discount in discounts_after:
            if discount["voucher"] and discount["voucher"] == voucher:
                found_discount = True
                break
        if not found_discount:
            messages.warning(
                self.request, _("Your basket does not qualify for a voucher discount")
            )
            self.request.basket.vouchers.remove(voucher)
        else:
            messages.info(
                self.request,
                _("Voucher '%(code)s' added to basket") % {"code": voucher.code},
            )

    def form_valid(self, form):
        code = form.cleaned_data["code"]
        if not self.request.basket.id:
            return redirect_to_referrer(self.request, "basket:summary")
        if self.request.basket.contains_voucher(code):
            messages.error(
                self.request,
                _("You have already added the '%(code)s' voucher to your basket")
                % {"code": code},
            )
        else:
            try:
                voucher = self.voucher_model._default_manager.get(code=code)
            except self.voucher_model.DoesNotExist:
                messages.error(
                    self.request,
                    _("No voucher found with code '%(code)s'") % {"code": code},
                )
            else:
                self.apply_voucher_to_basket(voucher)
        return redirect_to_referrer(self.request, "basket:summary")

    def form_invalid(self, form):
        messages.error(self.request, _("Please enter a voucher code"))
        return redirect(reverse("basket:summary") + "#voucher")


class VoucherRemoveView(View):
    voucher_model = get_model("voucher", "voucher")
    remove_signal = voucher_removal
    http_method_names = ["post"]

    def post(self, request, *args, **kwargs):
        response = redirect("basket:summary")

        voucher_id = kwargs["pk"]
        if not request.basket.id:
            # Hacking attempt - the basket must be saved for it to have
            # a voucher in it.
            return response
        try:
            voucher = request.basket.vouchers.get(id=voucher_id)
        except ObjectDoesNotExist:
            messages.error(request, _("No voucher found with id '%s'") % voucher_id)
        else:
            request.basket.vouchers.remove(voucher)
            self.remove_signal.send(sender=self, basket=request.basket, voucher=voucher)
            messages.info(request, _("Voucher '%s' removed from basket") % voucher.code)

        return response


class SavedView(ModelFormSetView):
    model = get_model("basket", "line")
    basket_model = get_model("basket", "basket")
    formset_class = SavedLineFormSet
    form_class = SavedLineForm
    factory_kwargs = {"extra": 0, "can_delete": True}

    def get(self, request, *args, **kwargs):
        return redirect("basket:summary")

    def get_queryset(self):
        """
        Return list of :py:class:`Line <oscar.apps.basket.abstract_models.AbstractLine>`
        instances associated with the saved basked associated with the currently
        authenticated user.
        """
        try:
            saved_basket = self.basket_model.saved.get(owner=self.request.user)
            saved_basket.strategy = self.request.strategy
            return saved_basket.all_lines()
        except self.basket_model.DoesNotExist:
            return []

    def get_success_url(self):
        return safe_referrer(self.request, "basket:summary")

    def get_formset_kwargs(self):
        kwargs = super().get_formset_kwargs()
        kwargs["prefix"] = "saved"
        kwargs["basket"] = self.request.basket
        kwargs["strategy"] = self.request.strategy
        return kwargs

    def formset_valid(self, formset):
        offers_before = self.request.basket.applied_offers()

        is_move = False
        for form in formset:
            if form.cleaned_data.get("move_to_basket", False):
                is_move = True
                msg = render_to_string(
                    "oscar/basket/messages/line_restored.html", {"line": form.instance}
                )
                messages.info(self.request, msg, extra_tags="safe noicon")
                real_basket = self.request.basket
                real_basket.merge_line(form.instance)

        if is_move:
            # As we're changing the basket, we need to check if it qualifies
            # for any new offers.
            BasketMessageGenerator().apply_messages(self.request, offers_before)
            response = redirect(self.get_success_url())
        else:
            response = super().formset_valid(formset)
        return response

    def formset_invalid(self, formset):
        messages.error(
            self.request,
            "\n".join(
                error for ed in formset.errors for el in ed.values() for error in el
            ),
        )
        return redirect_to_referrer(self.request, "basket:summary")



from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Count, Sum
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

from oscar.apps.catalogue.reviews.utils import get_default_review_status
from oscar.core import validators
from oscar.core.compat import AUTH_USER_MODEL
from oscar.core.loading import get_class

ProductReviewQuerySet = get_class("catalogue.reviews.managers", "ProductReviewQuerySet")


class AbstractProductReview(models.Model):
    """
    A review of a product

    Reviews can belong to a user or be anonymous.
    """

    product = models.ForeignKey(
        "catalogue.Product", related_name="reviews", null=True, on_delete=models.CASCADE
    )

    # Scores are between 0 and 5
    SCORE_CHOICES = tuple([(x, x) for x in range(0, 6)])
    score = models.SmallIntegerField(_("Score"), choices=SCORE_CHOICES)

    title = models.CharField(
        verbose_name=pgettext_lazy("Product review title", "Title"),
        max_length=255,
        validators=[validators.non_whitespace],
    )

    body = models.TextField(_("Body"))

    # User information.
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="reviews",
    )

    # Fields to be completed if user is anonymous
    name = models.CharField(
        pgettext_lazy("Anonymous reviewer name", "Name"), max_length=255, blank=True
    )
    email = models.EmailField(_("Email"), blank=True)
    homepage = models.URLField(_("URL"), blank=True)

    FOR_MODERATION, APPROVED, REJECTED = 0, 1, 2
    STATUS_CHOICES = (
        (FOR_MODERATION, _("Requires moderation")),
        (APPROVED, _("Approved")),
        (REJECTED, _("Rejected")),
    )

    status = models.SmallIntegerField(
        _("Status"), choices=STATUS_CHOICES, default=get_default_review_status
    )

    # Denormalised vote totals
    total_votes = models.IntegerField(
        _("Total Votes"), default=0
    )  # upvotes + down votes
    delta_votes = models.IntegerField(
        _("Delta Votes"), default=0, db_index=True
    )  # upvotes - down votes

    date_created = models.DateTimeField(auto_now_add=True)

    # Managers
    objects = ProductReviewQuerySet.as_manager()

    class Meta:
        abstract = True
        app_label = "reviews"
        ordering = ["-delta_votes", "id"]
        unique_together = (("product", "user"),)
        verbose_name = _("Product review")
        verbose_name_plural = _("Product reviews")

    def get_absolute_url(self):
        kwargs = {
            "product_slug": self.product.slug,
            "product_pk": self.product.id,
            "pk": self.id,
        }
        return reverse("catalogue:reviews-detail", kwargs=kwargs)

    def __str__(self):
        return self.title

    def clean(self):
        self.title = self.title.strip()
        self.body = self.body.strip()
        if not self.user and not (self.name and self.email):
            raise ValidationError(
                _("Anonymous reviews must include a name and an email")
            )

    def vote_up(self, user):
        self.votes.create(user=user, delta=AbstractVote.UP)

    def vote_down(self, user):
        self.votes.create(user=user, delta=AbstractVote.DOWN)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.product.update_rating()

    def delete(self, *args, **kwargs):
        super().delete(*args, **kwargs)
        if self.product is not None:
            self.product.update_rating()

    # Properties

    @property
    def is_anonymous(self):
        return self.user is None

    @property
    def pending_moderation(self):
        return self.status == self.FOR_MODERATION

    @property
    def is_approved(self):
        return self.status == self.APPROVED

    @property
    def is_rejected(self):
        return self.status == self.REJECTED

    @property
    def has_votes(self):
        return self.total_votes > 0

    @property
    def num_up_votes(self):
        """Returns the total up votes"""
        return int((self.total_votes + self.delta_votes) / 2)

    @property
    def num_down_votes(self):
        """Returns the total down votes"""
        return int((self.total_votes - self.delta_votes) / 2)

    @property
    def reviewer_name(self):
        if self.user:
            name = self.user.get_full_name()
            return name if name else _("anonymous")
        else:
            return self.name

    # Helpers

    def update_totals(self):
        """
        Update total and delta votes
        """
        result = self.votes.aggregate(score=Sum("delta"), total_votes=Count("id"))
        self.total_votes = result["total_votes"] or 0
        self.delta_votes = result["score"] or 0
        self.save()

    def can_user_vote(self, user):
        """
        Test whether the passed user is allowed to vote on this
        review
        """
        if not user.is_authenticated:
            return False, _("Only signed in users can vote")
        # pylint: disable=no-member
        vote = self.votes.model(review=self, user=user, delta=1)
        try:
            vote.full_clean()
        except ValidationError as e:
            return False, "%s" % e
        return True, ""


class AbstractVote(models.Model):
    """
    Records user ratings as yes/no vote.

    * Only signed-in users can vote.
    * Each user can vote only once.
    """

    review = models.ForeignKey(
        "reviews.ProductReview", on_delete=models.CASCADE, related_name="votes"
    )
    user = models.ForeignKey(
        AUTH_USER_MODEL, related_name="review_votes", on_delete=models.CASCADE
    )
    UP, DOWN = 1, -1
    VOTE_CHOICES = ((UP, _("Up")), (DOWN, _("Down")))
    delta = models.SmallIntegerField(_("Delta"), choices=VOTE_CHOICES)
    date_created = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True
        app_label = "reviews"
        ordering = ["-date_created"]
        unique_together = (("user", "review"),)
        verbose_name = _("Vote")
        verbose_name_plural = _("Votes")

    def __str__(self):
        return "%s vote for %s" % (self.delta, self.review)

    def clean(self):
        if not self.review.is_anonymous and self.review.user == self.user:
            raise ValidationError(_("You cannot vote on your own reviews"))
        if not self.user.id:
            raise ValidationError(_("Only signed-in users can vote on reviews"))
        previous_votes = self.review.votes.filter(user=self.user)
        if len(previous_votes) > 0:
            raise ValidationError(_("You can only vote once on a review"))

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.review.update_totals()



from django.utils.translation import gettext_lazy as _

from oscar.core.loading import get_class
from oscar.apps.search.signals import user_search

SearchForm = get_class("search.forms", "SearchForm")
BaseSearchView = get_class("search.views.base", "BaseSearchView")


class FacetedSearchView(BaseSearchView):
    form_class = SearchForm
    template_name = "oscar/search/results.html"
    context_object_name = "results"

    def dispatch(self, request, *args, **kwargs):
        # Raise a signal for other apps to hook into for analytics
        user_search.send(
            sender=self,
            session=self.request.session,
            user=self.request.user,
            query=self.request.GET.get("q", ""),
        )

        return super().dispatch(request, *args, **kwargs)



# Generated by Django 3.2.15 on 2022-09-28 13:34
import json
from ast import literal_eval

from django.db import migrations
from django.core.serializers.json import DjangoJSONEncoder


def forward(apps, schema_editor):
    LineAttribute = apps.get_model("basket", "LineAttribute")
    for at in LineAttribute.objects.all():
        try:  # if the value is allready valid json, continue
            json.loads(at.value)
            continue
        except json.JSONDecodeError:
            pass

        try:  # to parse the string a python, then convert to json, then continue
            val = literal_eval(at.value)
            at.value = json.dumps(val, cls=DjangoJSONEncoder)
            at.save()
            continue
        except (ValueError, SyntaxError):
            pass

        # convert the string to json as it is
        at.value = json.dumps(at.value)
        at.save()


class Migration(migrations.Migration):

    replaces = [
        ('basket', '0010_json_basket_option')
    ]

    dependencies = [
        ('basket', '0009_line_date_updated'),
    ]

    operations = [
        migrations.RunPython(forward, migrations.RunPython.noop),
    ]



from django.core.management.base import BaseCommand, CommandError

from oscar.core.loading import get_model

Order = get_model("order", "Order")
CommunicationEventType = get_model("communication", "CommunicationEventType")


class Command(BaseCommand):
    help = "For testing the content of order emails"

    def add_arguments(self, parser):
        parser.add_argument("event_type", help="The CommunicationEventType")
        parser.add_argument("order_number", help="The Order number")

    def handle(self, *args, **options):
        try:
            order = Order.objects.get(number=options["order_number"])
        except Order.DoesNotExist:
            raise CommandError(
                "No order found with number %s" % options["order_number"]
            )

        ctx = {
            "order": order,
            "lines": order.lines.all(),
        }
        messages = CommunicationEventType.objects.get_and_render(
            options["event_type"], ctx
        )
        print(
            "Subject: %s\nBody:\n\n%s\nBody HTML:\n\n%s"
            % (messages["subject"], messages["body"], messages["html"])
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.utils.module_loading import import_string
from django.conf import settings
from decimal import Decimal

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ProductRecord',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('num_views', models.PositiveIntegerField(default=0, verbose_name='Views')),
                ('num_basket_additions', models.PositiveIntegerField(default=0, verbose_name='Basket Additions')),
                ('num_purchases', models.PositiveIntegerField(default=0, db_index=True, verbose_name='Purchases')),
                ('score', models.FloatField(default=0.0, verbose_name='Score')),
            ],
            options={
                'ordering': ['-num_purchases'],
                'verbose_name_plural': 'Product records',
                'verbose_name': 'Product record',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='UserProductView',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
            ],
            options={
                'verbose_name_plural': 'User product views',
                'verbose_name': 'User product view',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='UserRecord',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('num_product_views', models.PositiveIntegerField(default=0, verbose_name='Product Views')),
                ('num_basket_additions', models.PositiveIntegerField(default=0, verbose_name='Basket Additions')),
                ('num_orders', models.PositiveIntegerField(default=0, db_index=True, verbose_name='Orders')),
                ('num_order_lines', models.PositiveIntegerField(default=0, db_index=True, verbose_name='Order Lines')),
                ('num_order_items', models.PositiveIntegerField(default=0, db_index=True, verbose_name='Order Items')),
                ('total_spent', models.DecimalField(default=Decimal('0.00'), max_digits=12, decimal_places=2, verbose_name='Total Spent')),
                ('date_last_order', models.DateTimeField(blank=True, verbose_name='Last Order Date', null=True)),
                ('user', models.OneToOneField(verbose_name='User', to=settings.AUTH_USER_MODEL,
                                              on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'User records',
                'verbose_name': 'User record',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='UserSearch',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('query', models.CharField(max_length=255, db_index=True, verbose_name='Search term')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('user', models.ForeignKey(verbose_name='User', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'User search queries',
                'verbose_name': 'User search query',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
    ]



from http import client as http_client

from django.contrib.auth.models import Permission
from django.urls import reverse
from django_webtest import WebTest
from purl import URL

from oscar.core.compat import get_user_model

User = get_user_model()


def add_permissions(user, permissions):
    """
    Grant permissions to the passed user

    :param permissions: e.g. ['partner.dashboard_access']
    """
    for permission in permissions:
        app_label, __, codename = permission.partition(".")
        if codename:
            perm = Permission.objects.get(
                content_type__app_label=app_label, codename=codename
            )
            user.user_permissions.add(perm)


class WebTestCase(WebTest):
    is_staff = False
    is_anonymous = False
    is_superuser = False

    username = "testuser"
    email = "testuser@buymore.com"
    password = "somefancypassword"
    permissions = []

    def setUp(self):
        self.user = None

        if not self.is_anonymous:
            self.user = self.create_user(self.username, self.email, self.password)
            self.user.is_staff = self.is_staff
            add_permissions(self.user, self.permissions)
            self.user.save()

    def create_user(self, username=None, email=None, password=None):
        """
        Create a user for use in a test.

        As usernames are optional in newer versions of Django, it only sets it
        if exists.
        """
        kwargs = {"email": email, "password": password}
        fields = {f.name: f for f in User._meta.get_fields()}

        if "username" in fields:
            kwargs["username"] = username
        return User.objects.create_user(**kwargs)

    def get(self, url, **kwargs):
        kwargs.setdefault("user", self.user)
        return self.app.get(url, **kwargs)

    def post(self, url, **kwargs):
        kwargs.setdefault("user", self.user)
        return self.app.post(url, **kwargs)

    # Custom assertions

    def assertIsRedirect(self, response, expected_url=None):
        self.assertTrue(
            response.status_code in (http_client.FOUND, http_client.MOVED_PERMANENTLY)
        )
        if expected_url:
            location = URL.from_string(response["Location"])
            self.assertEqual(expected_url, location.path())

    def assertIsNotRedirect(self, response):
        self.assertIsOk(response)
        self.assertTrue(
            response.status_code
            not in (http_client.FOUND, http_client.MOVED_PERMANENTLY)
        )

    def assertRedirectsTo(self, response, url_name, kwargs=None):
        """
        Asserts that a response is a redirect to a given URL name.
        """
        self.assertIsRedirect(response)
        location = response.headers["Location"]
        for unwanted in ["http://localhost:80", "http://testserver"]:
            location = location.replace(unwanted, "")
        self.assertEqual(reverse(url_name, kwargs=kwargs), location)

    def assertNoAccess(self, response, msg=None):
        self.assertContext(response)
        self.assertTrue(
            response.status_code in (http_client.NOT_FOUND, http_client.FORBIDDEN), msg
        )

    def assertIsOk(self, response, msg=None):
        self.assertEqual(http_client.OK, response.status_code, msg)

    def assertContext(self, response):
        self.assertTrue(response.context is not None, "No context was returned")

    def assertInContext(self, response, key):
        self.assertContext(response)
        self.assertTrue(
            key in response.context, "Context should contain a variable '%s'" % key
        )

    def assertNotInContext(self, response, key):
        self.assertContext(response)
        self.assertTrue(
            key not in response.context,
            "Context should not contain a variable '%s'" % key,
        )



from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.urls import path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig
from oscar.core.loading import get_class


class CheckoutConfig(OscarConfig):
    label = "checkout"
    name = "oscar.apps.checkout"
    verbose_name = _("Checkout")

    namespace = "checkout"

    # pylint: disable=attribute-defined-outside-init
    def ready(self):
        self.index_view = get_class("checkout.views", "IndexView")
        self.shipping_address_view = get_class("checkout.views", "ShippingAddressView")
        self.user_address_update_view = get_class(
            "checkout.views", "UserAddressUpdateView"
        )
        self.user_address_delete_view = get_class(
            "checkout.views", "UserAddressDeleteView"
        )
        self.shipping_method_view = get_class("checkout.views", "ShippingMethodView")
        self.payment_method_view = get_class("checkout.views", "PaymentMethodView")
        self.payment_details_view = get_class("checkout.views", "PaymentDetailsView")
        self.thankyou_view = get_class("checkout.views", "ThankYouView")

    def get_urls(self):
        urls = [
            path("", self.index_view.as_view(), name="index"),
            # Shipping/user address views
            path(
                "shipping-address/",
                self.shipping_address_view.as_view(),
                name="shipping-address",
            ),
            path(
                "user-address/edit/<int:pk>/",
                self.user_address_update_view.as_view(),
                name="user-address-update",
            ),
            path(
                "user-address/delete/<int:pk>/",
                self.user_address_delete_view.as_view(),
                name="user-address-delete",
            ),
            # Shipping method views
            path(
                "shipping-method/",
                self.shipping_method_view.as_view(),
                name="shipping-method",
            ),
            # Payment views
            path(
                "payment-method/",
                self.payment_method_view.as_view(),
                name="payment-method",
            ),
            path(
                "payment-details/",
                self.payment_details_view.as_view(),
                name="payment-details",
            ),
            # Preview and thankyou
            path(
                "preview/",
                self.payment_details_view.as_view(preview=True),
                name="preview",
            ),
            path("thank-you/", self.thankyou_view.as_view(), name="thank-you"),
        ]
        return self.post_process_urls(urls)

    def get_url_decorator(self, pattern):
        if not settings.OSCAR_ALLOW_ANON_CHECKOUT:
            return login_required
        if pattern.name.startswith("user-address"):
            return login_required
        return None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django.db.models.deletion
from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)


class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Line',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.PositiveIntegerField(default=1, verbose_name='Quantity')),
                ('title', models.CharField(max_length=255, verbose_name='Title')),
                ('product', models.ForeignKey(null=True, verbose_name='Product', on_delete=django.db.models.deletion.SET_NULL, related_name='wishlists_lines', to='catalogue.Product', blank=True)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Wish list line',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='WishList',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(default='Default', max_length=255, verbose_name='Name')),
                ('key', models.CharField(max_length=6, unique=True, db_index=True, verbose_name='Key', editable=False)),
                ('visibility', models.CharField(default='Private', max_length=20, verbose_name='Visibility', choices=[('Private', 'Private - Only the owner can see the wish list'), ('Shared', 'Shared - Only the owner and people with access to the obfuscated link can see the wish list'), ('Public', 'Public - Everybody can see the wish list')])),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date created')),
                ('owner', models.ForeignKey(verbose_name='Owner', related_name='wishlists', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
                'ordering': ('owner', 'date_created'),
                'abstract': False,
                'verbose_name': 'Wish List',
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='line',
            name='wishlist',
            field=models.ForeignKey(verbose_name='Wish List', related_name='lines', to='wishlists.WishList', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='line',
            unique_together=set([('wishlist', 'product')]),
        ),
    ]



# Generated by Django 4.2.9 on 2024-06-03 8:06
from django.db import migrations


def update_line_num_allocated(apps, schema_editor):
    Line = apps.get_model("order", "Line")
    for line in Line.objects.all():
        if line.num_allocated is None:
            line.num_allocated = line.quantity
            line.save(update_fields=["num_allocated"])


class Migration(migrations.Migration):
    dependencies = [
        ('order', '0016_line_allocation_cancelled_line_num_allocated'),
    ]

    operations = [
        migrations.RunPython(update_line_num_allocated, migrations.RunPython.noop),
    ]



# Generated by Django 4.2.6 on 2023-12-20 17:10

from django.db import migrations
import oscar.models.fields


class Migration(migrations.Migration):
    dependencies = [
        ("catalogue", "0026_predefined_product_options"),
    ]

    operations = [
        migrations.AddField(
            model_name="attributeoption",
            name="code",
            field=oscar.models.fields.NullCharField(
                max_length=255, unique=True, verbose_name="Unique identifier"
            ),
        ),
        migrations.AddField(
            model_name="attributeoptiongroup",
            name="code",
            field=oscar.models.fields.NullCharField(
                max_length=255, unique=True, verbose_name="Unique identifier"
            ),
        ),
        migrations.AddField(
            model_name="category",
            name="code",
            field=oscar.models.fields.NullCharField(
                max_length=255, unique=True, verbose_name="Code"
            ),
        ),
        migrations.AddField(
            model_name="productimage",
            name="code",
            field=oscar.models.fields.NullCharField(
                max_length=255, unique=True, verbose_name="Code"
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import oscar.models.fields.autoslugfield
import oscar.models.fields
from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)



class Migration(migrations.Migration):

    dependencies = [
        ('catalogue', '0001_initial'),
        ('address', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Partner',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', oscar.models.fields.autoslugfield.AutoSlugField(populate_from='name', unique=True, verbose_name='Code', max_length=128, editable=False, blank=True)),
                ('name', models.CharField(max_length=128, verbose_name='Name', blank=True)),
                ('users', models.ManyToManyField(related_name='partners', blank=True, verbose_name='Users', to=settings.AUTH_USER_MODEL, null=True)),
            ],
            options={
                'verbose_name_plural': 'Fulfillment partners',
                'verbose_name': 'Fulfillment partner',
                'abstract': False,
                'permissions': (('dashboard_access', 'Can access dashboard'),),
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PartnerAddress',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(verbose_name='Title', max_length=64, blank=True, choices=[('Mr', 'Mr'), ('Miss', 'Miss'), ('Mrs', 'Mrs'), ('Ms', 'Ms'), ('Dr', 'Dr')])),
                ('first_name', models.CharField(max_length=255, verbose_name='First name', blank=True)),
                ('last_name', models.CharField(max_length=255, verbose_name='Last name', blank=True)),
                ('line1', models.CharField(max_length=255, verbose_name='First line of address')),
                ('line2', models.CharField(max_length=255, verbose_name='Second line of address', blank=True)),
                ('line3', models.CharField(max_length=255, verbose_name='Third line of address', blank=True)),
                ('line4', models.CharField(max_length=255, verbose_name='City', blank=True)),
                ('state', models.CharField(max_length=255, verbose_name='State/County', blank=True)),
                ('postcode', oscar.models.fields.UppercaseCharField(max_length=64, verbose_name='Post/Zip-code', blank=True)),
                ('search_text', models.TextField(editable=False, verbose_name='Search text - used only for searching addresses')),
                ('country', models.ForeignKey(verbose_name='Country', to='address.Country', on_delete=models.CASCADE)),
                ('partner', models.ForeignKey(verbose_name='Partner', related_name='addresses', to='partner.Partner', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Partner addresses',
                'verbose_name': 'Partner address',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='StockAlert',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('threshold', models.PositiveIntegerField(verbose_name='Threshold')),
                ('status', models.CharField(default='Open', max_length=128, verbose_name='Status', choices=[('Open', 'Open'), ('Closed', 'Closed')])),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Created')),
                ('date_closed', models.DateTimeField(blank=True, verbose_name='Date Closed', null=True)),
            ],
            options={
                'ordering': ('-date_created',),
                'verbose_name_plural': 'Stock alerts',
                'verbose_name': 'Stock alert',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='StockRecord',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('partner_sku', models.CharField(max_length=128, verbose_name='Partner SKU')),
                ('price_currency', models.CharField(default='GBP', max_length=12, verbose_name='Currency')),
                ('price_excl_tax', models.DecimalField(max_digits=12, decimal_places=2, blank=True, verbose_name='Price (excl. tax)', null=True)),
                ('price_retail', models.DecimalField(max_digits=12, decimal_places=2, blank=True, verbose_name='Price (retail)', null=True)),
                ('cost_price', models.DecimalField(max_digits=12, decimal_places=2, blank=True, verbose_name='Cost Price', null=True)),
                ('num_in_stock', models.PositiveIntegerField(blank=True, verbose_name='Number in stock', null=True)),
                ('num_allocated', models.IntegerField(blank=True, verbose_name='Number allocated', null=True)),
                ('low_stock_threshold', models.PositiveIntegerField(blank=True, verbose_name='Low Stock Threshold', null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date created')),
                ('date_updated', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Date updated')),
                ('partner', models.ForeignKey(verbose_name='Partner', related_name='stockrecords', to='partner.Partner', on_delete=models.CASCADE)),
                ('product', models.ForeignKey(verbose_name='Product', related_name='stockrecords', to='catalogue.Product', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name_plural': 'Stock records',
                'verbose_name': 'Stock record',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='stockrecord',
            unique_together=set([('partner', 'partner_sku')]),
        ),
        migrations.AddField(
            model_name='stockalert',
            name='stockrecord',
            field=models.ForeignKey(verbose_name='Stock Record', related_name='alerts', to='partner.StockRecord', on_delete=models.CASCADE),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import oscar.models.fields.autoslugfield
from decimal import Decimal
import django.core.validators
from django.utils.module_loading import import_string
from django.conf import settings

models_AutoField = import_string(settings.DEFAULT_AUTO_FIELD)


class Migration(migrations.Migration):

    dependencies = [
        ('address', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='OrderAndItemCharges',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', oscar.models.fields.autoslugfield.AutoSlugField(populate_from='name', unique=True, verbose_name='Slug', max_length=128, editable=False, blank=True)),
                ('name', models.CharField(unique=True, max_length=128, verbose_name='Name')),
                ('description', models.TextField(verbose_name='Description', blank=True)),
                ('price_per_order', models.DecimalField(default=Decimal('0.00'), max_digits=12, decimal_places=2, verbose_name='Price per order')),
                ('price_per_item', models.DecimalField(default=Decimal('0.00'), max_digits=12, decimal_places=2, verbose_name='Price per item')),
                ('free_shipping_threshold', models.DecimalField(max_digits=12, decimal_places=2, blank=True, verbose_name='Free Shipping', null=True)),
                ('countries', models.ManyToManyField(blank=True, verbose_name='Countries', to='address.Country', null=True)),
            ],
            options={
                'ordering': ['name'],
                'verbose_name_plural': 'Order and Item Charges',
                'verbose_name': 'Order and Item Charge',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='WeightBand',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('upper_limit', models.DecimalField(verbose_name='Upper Limit', decimal_places=3, validators=[django.core.validators.MinValueValidator(Decimal('0.00'))], help_text='Enter upper limit of this weight band in kg. The lower limit will be determined by the other weight bands.', max_digits=12)),
                ('charge', models.DecimalField(max_digits=12, decimal_places=2, validators=[django.core.validators.MinValueValidator(Decimal('0.00'))], verbose_name='Charge')),
            ],
            options={
                'ordering': ['method', 'upper_limit'],
                'verbose_name_plural': 'Weight Bands',
                'verbose_name': 'Weight Band',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='WeightBased',
            fields=[
                ('id', models_AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', oscar.models.fields.autoslugfield.AutoSlugField(populate_from='name', unique=True, verbose_name='Slug', max_length=128, editable=False, blank=True)),
                ('name', models.CharField(unique=True, max_length=128, verbose_name='Name')),
                ('description', models.TextField(verbose_name='Description', blank=True)),
                ('default_weight', models.DecimalField(validators=[django.core.validators.MinValueValidator(Decimal('0.00'))], verbose_name='Default Weight', default=Decimal('0.000'), max_digits=12, decimal_places=3, help_text='Default product weight in kg when no weight attribute is defined')),
                ('countries', models.ManyToManyField(blank=True, verbose_name='Countries', to='address.Country', null=True)),
            ],
            options={
                'ordering': ['name'],
                'verbose_name_plural': 'Weight-based Shipping Methods',
                'verbose_name': 'Weight-based Shipping Method',
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='weightband',
            name='method',
            field=models.ForeignKey(verbose_name='Method', related_name='bands', to='shipping.WeightBased', on_delete=models.CASCADE),
            preserve_default=True,
        ),
    ]



# Generated by Django 3.2.9 on 2022-01-06 16:24

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('basket', '0010_convert_to_valid_json'),
    ]

    operations = [
        migrations.AlterField(
            model_name='lineattribute',
            name='value',
            field=models.JSONField(encoder=django.core.serializers.json.DjangoJSONEncoder, verbose_name='Value'),
        ),
    ]



import copy
import re

from django.forms import Widget
from django.urls import reverse


class RelatedFieldWidgetWrapper(Widget):
    """
    This class is a wrapper to a given widget to add the add icon for the
    Oscar dashboard.
    """

    template_name = "oscar/dashboard/widgets/related_widget_wrapper.html"

    IS_POPUP_VALUE = "1"
    IS_POPUP_VAR = "_popup"
    TO_FIELD_VAR = "_to_field"

    # pylint: disable=super-init-not-called
    def __init__(self, widget, rel):
        self.needs_multipart_form = widget.needs_multipart_form
        self.attrs = widget.attrs
        self.choices = widget.choices
        self.widget = widget
        self.rel = rel

    def __deepcopy__(self, memo):
        obj = copy.copy(self)
        obj.widget = copy.deepcopy(self.widget, memo)
        obj.attrs = self.widget.attrs
        memo[id(self)] = obj
        return obj

    @property
    def is_hidden(self):
        return self.widget.is_hidden

    @property
    def media(self):
        return self.widget.media

    def get_related_url(self, info, action, *args):
        app_label = info[0]
        model_object_name = info[1]
        # Convert the model's object name into lowercase, with dashes between
        # the camel-cased words
        model_object_name = "-".join(
            re.sub("([a-z])([A-Z])", r"\1 \2", model_object_name).lower().split()
        )
        # Does not specify current app
        return reverse(
            "dashboard:%s-%s-%s" % (app_label, model_object_name, action), args=args
        )

    def get_context(self, name, value, attrs):
        rel_opts = self.rel.model._meta
        info = (rel_opts.app_label, rel_opts.object_name)
        self.widget.choices = self.choices
        url_params = "&".join(
            "%s=%s" % param
            for param in [
                (
                    RelatedFieldWidgetWrapper.TO_FIELD_VAR,
                    self.rel.get_related_field().name,
                ),
                (
                    RelatedFieldWidgetWrapper.IS_POPUP_VAR,
                    RelatedFieldWidgetWrapper.IS_POPUP_VALUE,
                ),
            ]
        )
        context = {
            "rendered_widget": self.widget.render(name, value, attrs),
            "name": name,
            "url_params": url_params,
            "model": rel_opts.verbose_name,
        }
        change_related_template_url = self.get_related_url(info, "update", "__fk__")
        context.update(
            change_related_template_url=change_related_template_url,
        )
        add_related_url = self.get_related_url(info, "create")
        context.update(
            add_related_url=add_related_url,
        )
        delete_related_template_url = self.get_related_url(info, "delete", "__fk__")
        context.update(
            delete_related_template_url=delete_related_template_url,
        )
        return context

    def value_from_datadict(self, data, files, name):
        return self.widget.value_from_datadict(data, files, name)

    def value_omitted_from_data(self, data, files, name):
        return self.widget.value_omitted_from_data(data, files, name)

    def id_for_label(self, id_):
        return self.widget.id_for_label(id_)


class RelatedMultipleFieldWidgetWrapper(RelatedFieldWidgetWrapper):
    template_name = "oscar/dashboard/widgets/related_multiple_widget_wrapper.html"



from django import template

from oscar.core.loading import feature_hidden

register = template.Library()


@register.simple_tag(takes_context=True)
def get_parameters(context, except_field):
    """
    Renders current get parameters except for the specified parameter
    """
    getvars = context["request"].GET.copy()
    getvars.pop(except_field, None)
    if len(getvars.keys()) > 0:
        return "%s&" % getvars.urlencode()

    return ""


@register.tag()
def iffeature(parser, token):
    nodelist = parser.parse(("endiffeature",))
    try:
        (
            tag_name,
            app_name,
        ) = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError(
            "%r tag requires a single argument" % token.contents.split()[0]
        )
    if not (app_name[0] == app_name[-1] and app_name[0] in ('"', "'")):
        raise template.TemplateSyntaxError(
            "%r tag's argument should be in quotes" % tag_name
        )
    parser.delete_first_token()
    return ConditionalOutputNode(nodelist, app_name[1:-1])


class ConditionalOutputNode(template.Node):
    def __init__(self, nodelist, feature_name):
        self.nodelist = nodelist
        self.feature_name = feature_name

    def render(self, context):
        if not feature_hidden(self.feature_name):
            output = self.nodelist.render(context)
            return output
        else:
            return ""



import sys
import traceback
import warnings
from functools import lru_cache
from importlib import import_module

from django.apps import apps
from django.apps.config import MODELS_MODULE_NAME
from django.conf import settings
from django.core.exceptions import AppRegistryNotReady
from django.utils.module_loading import import_string

from oscar.core.exceptions import (
    AppNotFoundError,
    ClassNotFoundError,
)
from oscar.utils.deprecation import RemovedInOscar32Warning

# To preserve backwards compatibility of loading classes which moved
# from one Oscar module to another, we look into the dictionary below
# for the moved items during loading.
MOVED_MODELS = {}


def get_class(module_label, classname, module_prefix="oscar.apps"):
    """
    Dynamically import a single class from the given module.

    This is a simple wrapper around `get_classes` for the case of loading a
    single class.

    Args:
        module_label (str): Module label comprising the app label and the
            module name, separated by a dot.  For example, 'catalogue.forms'.
        classname (str): Name of the class to be imported.

    Returns:
        The requested class object or `None` if it can't be found
    """
    return get_classes(module_label, [classname], module_prefix)[0]


@lru_cache(maxsize=100)
def get_class_loader():
    return import_string(settings.OSCAR_DYNAMIC_CLASS_LOADER)


def get_classes(module_label, classnames, module_prefix="oscar.apps"):
    class_loader = get_class_loader()
    return class_loader(module_label, classnames, module_prefix)


def default_class_loader(module_label, classnames, module_prefix):
    """
    Dynamically import a list of classes from the given module.

    This works by looking up a matching app from the app registry,
    against the passed module label.  If the requested class can't be found in
    the matching module, then we attempt to import it from the corresponding
    core app.

    This is very similar to ``django.db.models.get_model`` function for
    dynamically loading models.  This function is more general though as it can
    load any class from the matching app, not just a model.

    Args:
        module_label (str): Module label comprising the app label and the
            module name, separated by a dot.  For example, 'catalogue.forms'.
        classname (str): Name of the class to be imported.

    Returns:
        The requested class object or ``None`` if it can't be found

    Examples:

        Load a single class:

        >>> get_class('dashboard.catalogue.forms', 'ProductForm')
        oscar.apps.dashboard.catalogue.forms.ProductForm

        Load a list of classes:

        >>> get_classes('dashboard.catalogue.forms',
        ...             ['ProductForm', 'StockRecordForm'])
        [oscar.apps.dashboard.catalogue.forms.ProductForm,
         oscar.apps.dashboard.catalogue.forms.StockRecordForm]

    Raises:

        AppNotFoundError: If no app is found in ``INSTALLED_APPS`` that matches
            the passed module label.

        ImportError: If the attempted import of a class raises an
            ``ImportError``, it is re-raised
    """

    if "." not in module_label:
        # Importing from top-level modules is not supported, e.g.
        # get_class('shipping', 'Scale'). That should be easy to fix,
        # but @maikhoepfel had a stab and could not get it working reliably.
        # Overridable classes in a __init__.py might not be a good idea anyway.
        raise ValueError("Importing from top-level modules is not supported")

    # import from Oscar package (should succeed in most cases)
    # e.g. 'oscar.apps.dashboard.catalogue.forms'
    oscar_module_label = "%s.%s" % (module_prefix, module_label)
    oscar_module = _import_module(oscar_module_label, classnames)

    # returns e.g. 'oscar.apps.dashboard.catalogue',
    # 'yourproject.apps.dashboard.catalogue' or 'dashboard.catalogue',
    # depending on what is set in INSTALLED_APPS
    app_name = _find_registered_app_name(module_label)
    if app_name.startswith("%s." % module_prefix):
        # The entry is obviously an Oscar one, we don't import again
        local_module = None
    else:
        # Attempt to import the classes from the local module
        # e.g. 'yourproject.dashboard.catalogue.forms'
        local_module_label = ".".join(app_name.split(".") + module_label.split(".")[1:])
        local_module = _import_module(local_module_label, classnames)

    if oscar_module is local_module is None:
        # This intentionally doesn't raise an ImportError, because ImportError
        # can get masked in complex circular import scenarios.
        raise ModuleNotFoundError(
            "The module with label '%s' could not be imported. This either"
            "means that it indeed does not exist, or you might have a problem"
            " with a circular import." % module_label
        )

    # return imported classes, giving preference to ones from the local package
    return _pluck_classes([local_module, oscar_module], classnames)


def _import_module(module_label, classnames):
    """
    Imports the module with the given name.
    Returns None if the module doesn't exist, but propagates any import errors.
    """
    try:
        return __import__(module_label, fromlist=classnames)
    except ImportError:
        # There are 2 reasons why there could be an ImportError:
        #
        #  1. Module does not exist. In that case, we ignore the import and
        #     return None
        #  2. Module exists but another ImportError occurred when trying to
        #     import the module. In that case, it is important to propagate the
        #     error.
        #
        # ImportError does not provide easy way to distinguish those two cases.
        # Fortunately, the traceback of the ImportError starts at __import__
        # statement. If the traceback has more than one frame, it means that
        # application was found and ImportError originates within the local app
        __, __, exc_traceback = sys.exc_info()
        frames = traceback.extract_tb(exc_traceback)
        if len(frames) > 1:
            raise


def _pluck_classes(modules, classnames):
    """
    Gets a list of class names and a list of modules to pick from.
    For each class name, will return the class from the first module that has a
    matching class.
    """
    klasses = []
    for classname in classnames:
        klass = None
        for module in modules:
            if hasattr(module, classname):
                klass = getattr(module, classname)
                break
        if not klass:
            packages = [m.__name__ for m in modules if m is not None]
            raise ClassNotFoundError(
                "No class '%s' found in %s" % (classname, ", ".join(packages))
            )
        klasses.append(klass)
    return klasses


def _find_registered_app_name(module_label):
    """
    Given a module label, finds the name of the matching Oscar app from the
    Django app registry.
    """
    from oscar.core.application import OscarConfig

    app_label = module_label.split(".")[0]
    try:
        app_config = apps.get_app_config(app_label)
    except LookupError:
        raise AppNotFoundError("Couldn't find an app to import %s from" % module_label)
    if not isinstance(app_config, OscarConfig):
        raise AppNotFoundError(
            "Couldn't find an Oscar app to import %s from" % module_label
        )
    return app_config.name


def get_profile_class():
    """
    Return the profile model class
    """
    # The AUTH_PROFILE_MODULE setting was deprecated in Django 1.5, but it
    # makes sense for Oscar to continue to use it. Projects built on Django
    # 1.4 are likely to have used a profile class and it's very difficult to
    # upgrade to a single user model. Hence, we should continue to support
    # having a separate profile class even if Django doesn't.
    setting = getattr(settings, "AUTH_PROFILE_MODULE", None)
    if setting is None:
        return None
    app_label, model_name = settings.AUTH_PROFILE_MODULE.split(".")
    return get_model(app_label, model_name)


def feature_hidden(feature_name):
    """
    Test if a certain Oscar feature is disabled.
    """
    return feature_name is not None and feature_name in settings.OSCAR_HIDDEN_FEATURES


def get_model(app_label, model_name):
    """
    Fetches a Django model using the app registry.

    This doesn't require that an app with the given app label exists,
    which makes it safe to call when the registry is being populated.
    All other methods to access models might raise an exception about the
    registry not being ready yet.
    Raises LookupError if model isn't found.
    """
    oscar_moved_model = MOVED_MODELS.get(app_label, None)
    if oscar_moved_model:
        if model_name.lower() in oscar_moved_model[1]:
            original_app_label = app_label
            app_label = oscar_moved_model[0]
            warnings.warn(
                "Model %s has recently moved from %s to the application %s, "
                "please update your imports."
                % (model_name, original_app_label, app_label),
                RemovedInOscar32Warning,
                stacklevel=2,
            )
    try:
        return apps.get_model(app_label, model_name)
    except AppRegistryNotReady:
        if apps.apps_ready and not apps.models_ready:
            # If this function is called while `apps.populate()` is
            # loading models, ensure that the module that defines the
            # target model has been imported and try looking the model up
            # in the app registry. This effectively emulates
            # `from path.to.app.models import Model` where we use
            # `Model = get_model('app', 'Model')` instead.
            app_config = apps.get_app_config(app_label)
            # `app_config.import_models()` cannot be used here because it
            # would interfere with `apps.populate()`.
            import_module("%s.%s" % (app_config.name, MODELS_MODULE_NAME))
            # In order to account for case-insensitivity of model_name,
            # look up the model through a private API of the app registry.
            return apps.get_registered_model(app_label, model_name)
        else:
            # This must be a different case (e.g. the model really doesn't
            # exist). We just re-raise the exception.
            raise


def is_model_registered(app_label, model_name):
    """
    Checks whether a given model is registered. This is used to only
    register Oscar models if they aren't overridden by a forked app.
    """
    try:
        apps.get_registered_model(app_label, model_name)
    except LookupError:
        return False
    else:
        return True


@lru_cache(maxsize=128)
def cached_import_string(path):
    return import_string(path)



# Generated by Django 3.0.9 on 2020-08-28 09:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('offer', '0009_auto_20200801_0817'),
    ]

    operations = [
        migrations.AddField(
            model_name='conditionaloffer',
            name='combinations',
            field=models.ManyToManyField(blank=True, help_text='Select other non-exclusive offers that this offer can be combined with on the same items', limit_choices_to={'exclusive': False}, related_name='in_combination', to='offer.ConditionalOffer'),
        ),
    ]



from django.contrib.auth import models as auth_models
from django.core.mail import send_mail
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _

from oscar.core.compat import AUTH_USER_MODEL


class UserManager(auth_models.BaseUserManager):
    def create_user(self, email, password=None, **extra_fields):
        """
        Creates and saves a User with the given email and
        password.
        """
        now = timezone.now()
        if not email:
            raise ValueError("The given email must be set")
        email = UserManager.normalize_email(email)
        user = self.model(
            email=email,
            is_staff=False,
            is_active=True,
            is_superuser=False,
            last_login=now,
            date_joined=now,
            **extra_fields,
        )

        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_superuser(self, email, password, **extra_fields):
        u = self.create_user(email, password, **extra_fields)
        u.is_staff = True
        u.is_active = True
        u.is_superuser = True
        u.save(using=self._db)
        return u


class AbstractUser(auth_models.AbstractBaseUser, auth_models.PermissionsMixin):
    """
    An abstract base user suitable for use in Oscar projects.

    This is basically a copy of the core AbstractUser model but without a
    username field
    """

    email = models.EmailField(_("email address"), unique=True)
    first_name = models.CharField(_("First name"), max_length=255, blank=True)
    last_name = models.CharField(_("Last name"), max_length=255, blank=True)
    is_staff = models.BooleanField(
        _("Staff status"),
        default=False,
        help_text=_("Designates whether the user can log into this admin site."),
    )
    is_active = models.BooleanField(
        _("Active"),
        default=True,
        help_text=_(
            "Designates whether this user should be treated as "
            "active. Unselect this instead of deleting accounts."
        ),
    )
    date_joined = models.DateTimeField(_("date joined"), default=timezone.now)

    objects = UserManager()

    USERNAME_FIELD = "email"

    class Meta:
        abstract = True
        verbose_name = _("User")
        verbose_name_plural = _("Users")

    def clean(self):
        super().clean()
        self.email = self.__class__.objects.normalize_email(self.email)

    def get_full_name(self):
        """
        Return the first_name plus the last_name, with a space in between.
        """
        full_name = "%s %s" % (self.first_name, self.last_name)
        return full_name.strip()

    def get_short_name(self):
        """
        Return the short name for the user.
        """
        return self.first_name

    def email_user(self, subject, message, from_email=None, **kwargs):
        """
        Send an email to this user.
        """
        send_mail(subject, message, from_email, [self.email], **kwargs)

    def _migrate_alerts_to_user(self):
        """
        Transfer any active alerts linked to a user's email address to the
        newly registered user.
        """
        # pylint: disable=no-member
        ProductAlert = self.alerts.model
        alerts = ProductAlert.objects.filter(
            email=self.email, status=ProductAlert.ACTIVE
        )
        alerts.update(user=self, key="", email="")

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        # Migrate any "anonymous" product alerts to the registered user
        # Ideally, this would be done via a post-save signal. But we can't
        # use get_user_model to wire up signals to custom user models
        # see Oscar ticket #1127, Django ticket #19218
        self._migrate_alerts_to_user()


class AbstractProductAlert(models.Model):
    """
    An alert for when a product comes back in stock
    """

    product = models.ForeignKey("catalogue.Product", on_delete=models.CASCADE)

    # A user is only required if the notification is created by a
    # registered user, anonymous users will only have an email address
    # attached to the notification
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.CASCADE,
        related_name="alerts",
        verbose_name=_("User"),
    )
    email = models.EmailField(_("Email"), db_index=True, blank=True)

    # This key are used to confirm and cancel alerts for anon users
    key = models.CharField(_("Key"), max_length=128, blank=True, db_index=True)

    # An alert can have two different statuses for authenticated
    # users ``ACTIVE`` and ``CANCELLED`` and anonymous users have an
    # additional status ``UNCONFIRMED``. For anonymous users a confirmation
    # and unsubscription key are generated when an instance is saved for
    # the first time and can be used to confirm and unsubscribe the
    # notifications.
    UNCONFIRMED, ACTIVE, CANCELLED, CLOSED = (
        "Unconfirmed",
        "Active",
        "Cancelled",
        "Closed",
    )
    STATUS_CHOICES = (
        (UNCONFIRMED, _("Not yet confirmed")),
        (ACTIVE, _("Active")),
        (CANCELLED, _("Cancelled")),
        (CLOSED, _("Closed")),
    )
    status = models.CharField(
        _("Status"), max_length=20, choices=STATUS_CHOICES, default=ACTIVE
    )

    date_created = models.DateTimeField(
        _("Date created"), auto_now_add=True, db_index=True
    )
    date_confirmed = models.DateTimeField(_("Date confirmed"), blank=True, null=True)
    date_cancelled = models.DateTimeField(_("Date cancelled"), blank=True, null=True)
    date_closed = models.DateTimeField(_("Date closed"), blank=True, null=True)

    class Meta:
        abstract = True
        app_label = "customer"
        ordering = ["-date_created"]
        verbose_name = _("Product alert")
        verbose_name_plural = _("Product alerts")

    @property
    def is_anonymous(self):
        return self.user is None

    @property
    def can_be_confirmed(self):
        return self.status == self.UNCONFIRMED

    @property
    def can_be_cancelled(self):
        return self.status in (self.ACTIVE, self.UNCONFIRMED)

    @property
    def is_cancelled(self):
        return self.status == self.CANCELLED

    @property
    def is_active(self):
        return self.status == self.ACTIVE

    def confirm(self):
        self.status = self.ACTIVE
        self.date_confirmed = timezone.now()
        self.save()

    confirm.alters_data = True

    def cancel(self):
        self.status = self.CANCELLED
        self.date_cancelled = timezone.now()
        self.save()

    cancel.alters_data = True

    def close(self):
        self.status = self.CLOSED
        self.date_closed = timezone.now()
        self.save()

    close.alters_data = True

    def get_email_address(self):
        if self.user:
            return self.user.email
        else:
            return self.email

    def save(self, *args, **kwargs):
        if not self.id and not self.user:
            self.key = self.get_random_key()
            self.status = self.UNCONFIRMED
        # Ensure date fields get updated when saving from modelform (which just
        # calls save, and doesn't call the methods cancel(), confirm() etc).
        if self.status == self.CANCELLED and self.date_cancelled is None:
            self.date_cancelled = timezone.now()
        if not self.user and self.status == self.ACTIVE and self.date_confirmed is None:
            self.date_confirmed = timezone.now()
        if self.status == self.CLOSED and self.date_closed is None:
            self.date_closed = timezone.now()

        return super().save(*args, **kwargs)

    def get_random_key(self):
        return get_random_string(
            length=40, allowed_chars="abcdefghijklmnopqrstuvwxyz0123456789"
        )

    def get_confirm_url(self):
        return reverse("customer:alerts-confirm", kwargs={"key": self.key})

    def get_cancel_url(self):
        return reverse("customer:alerts-cancel-by-key", kwargs={"key": self.key})



from django import template

register = template.Library()


@register.filter
def as_stars(value):
    """
    Convert a float rating between 0 and 5 to a CSS class

    The CSS class name is the number of stars to be displayed.

    * Rounds to the nearest integer
    * Maps no rating to 0 stars
    * Fails quietly
    """
    num_stars_to_class = {
        0: "",
        1: "One",
        2: "Two",
        3: "Three",
        4: "Four",
        5: "Five",
    }
    num_stars = int(round(value or 0.0))
    return num_stars_to_class.get(num_stars, "")


@register.filter
def may_vote(review, user):
    can_vote, __ = review.can_user_vote(user)
    return can_vote


@register.filter
def is_review_permitted(product, user):
    return product and product.is_review_permitted(user)



from decimal import Decimal as D

import factory
from django.conf import settings

from oscar.core.loading import get_class, get_model
from oscar.core.utils import slugify
from oscar.test import factories
from oscar.test.factories.utils import tax_add, tax_subtract

OrderCreator = get_class("order.utils", "OrderCreator")

__all__ = [
    "BillingAddressFactory",
    "ShippingAddressFactory",
    "OrderDiscountFactory",
    "OrderFactory",
    "OrderLineFactory",
    "ShippingEventTypeFactory",
    "ShippingEventFactory",
]


class BillingAddressFactory(factory.django.DjangoModelFactory):
    country = factory.SubFactory("oscar.test.factories.CountryFactory")

    first_name = "John"
    last_name = "Doe"
    line1 = "Streetname"
    line2 = "1a"
    line4 = "City"
    postcode = "1000AA"

    class Meta:
        model = get_model("order", "BillingAddress")


class ShippingAddressFactory(factory.django.DjangoModelFactory):
    country = factory.SubFactory("oscar.test.factories.CountryFactory")

    first_name = "John"
    last_name = "Doe"
    line1 = "Streetname"
    line2 = "1a"
    line4 = "City"
    postcode = "1000 AA"
    phone_number = "+12125555555"

    class Meta:
        model = get_model("order", "ShippingAddress")


class OrderDiscountFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = get_model("order", "OrderDiscount")


class OrderFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = get_model("order", "Order")

    if hasattr(settings, "OSCAR_INITIAL_ORDER_STATUS"):
        status = settings.OSCAR_INITIAL_ORDER_STATUS

    site_id = getattr(settings, "SITE_ID", None)
    number = factory.LazyAttribute(lambda o: "%d" % (100000 + o.basket.pk))
    basket = factory.SubFactory("oscar.test.factories.BasketFactory")

    shipping_code = "delivery"
    shipping_incl_tax = D("4.95")
    shipping_excl_tax = factory.LazyAttribute(
        lambda o: tax_subtract(o.shipping_incl_tax)
    )

    total_incl_tax = factory.LazyAttribute(lambda o: o.basket.total_incl_tax)
    total_excl_tax = factory.LazyAttribute(lambda o: o.basket.total_excl_tax)

    guest_email = factory.LazyAttribute(
        lambda o: (
            "%s.%s@example.com"
            % (o.billing_address.first_name[0], o.billing_address.last_name)
        ).lower()
    )

    shipping_address = factory.SubFactory(ShippingAddressFactory)
    billing_address = factory.SubFactory(BillingAddressFactory)

    @factory.post_generation
    def create_line_models(obj, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            if not obj.basket.all_lines().exists():
                product = factories.ProductFactory(stockrecords=None)
                factories.StockRecordFactory(
                    product=product, price_currency=settings.OSCAR_DEFAULT_CURRENCY
                )
                obj.basket.add_product(product)
            for line in obj.basket.all_lines():
                OrderCreator().create_line_models(obj, line)


class OrderLineFactory(factory.django.DjangoModelFactory):
    order = factory.SubFactory(OrderFactory)
    product = factory.SubFactory("oscar.test.factories.ProductFactory")
    partner_sku = factory.LazyAttribute(lambda x: x.product.upc)
    stockrecord = factory.LazyAttribute(lambda s: s.product.stockrecords.first())
    quantity = 1
    num_allocated = 1

    line_price_incl_tax = factory.LazyAttribute(
        lambda obj: tax_add(obj.stockrecord.price) * obj.quantity
    )
    line_price_excl_tax = factory.LazyAttribute(
        lambda obj: obj.stockrecord.price * obj.quantity
    )

    line_price_before_discounts_incl_tax = factory.SelfAttribute(".line_price_incl_tax")
    line_price_before_discounts_excl_tax = factory.SelfAttribute(".line_price_excl_tax")

    unit_price_incl_tax = factory.LazyAttribute(
        lambda obj: tax_add(obj.stockrecord.price)
    )
    unit_price_excl_tax = factory.LazyAttribute(lambda obj: obj.stockrecord.price)

    class Meta:
        model = get_model("order", "Line")


class ShippingEventTypeFactory(factory.django.DjangoModelFactory):
    name = "Test event"
    code = factory.LazyAttribute(lambda o: slugify(o.name).replace("-", "_"))

    class Meta:
        model = get_model("order", "ShippingEventType")
        django_get_or_create = ("code",)


class ShippingEventFactory(factory.django.DjangoModelFactory):
    order = factory.SubFactory(OrderFactory)
    event_type = factory.SubFactory(ShippingEventTypeFactory)

    class Meta:
        model = get_model("order", "ShippingEvent")



from django.conf import settings
from django.contrib import messages
from django.core.signing import BadSignature, Signer
from django.utils.functional import SimpleLazyObject, empty
from django.utils.translation import ngettext_lazy

from oscar.core.loading import get_class, get_model

Applicator = get_class("offer.applicator", "Applicator")
Basket = get_model("basket", "basket")
Selector = get_class("partner.strategy", "Selector")

selector = Selector()


class BasketMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Keep track of cookies that need to be deleted (which can only be done
        # when we're processing the response instance).
        request.cookies_to_delete = []

        # Load stock/price strategy and assign to request (it will later be
        # assigned to the basket too).
        strategy = selector.strategy(request=request, user=request.user)
        request.strategy = strategy

        # We lazily load the basket so use a private variable to hold the
        # cached instance.
        request._basket_cache = None

        def load_full_basket():
            """
            Return the basket after applying offers.
            """
            basket = self.get_basket(request)
            basket.strategy = request.strategy
            self.apply_offers_to_basket(request, basket)

            return basket

        def load_basket_hash():
            """
            Load the basket and return the basket hash

            Note that we don't apply offers or check that every line has a
            stockrecord here.
            """
            basket = self.get_basket(request)
            if basket.id:
                return self.get_basket_hash(basket.id)

        # Use Django's SimpleLazyObject to only perform the loading work
        # when the attribute is accessed.
        request.basket = SimpleLazyObject(load_full_basket)
        request.basket_hash = SimpleLazyObject(load_basket_hash)

        response = self.get_response(request)
        return self.process_response(request, response)

    def process_response(self, request, response):
        # Delete any surplus cookies
        cookies_to_delete = getattr(request, "cookies_to_delete", [])
        for cookie_key in cookies_to_delete:
            response.delete_cookie(cookie_key)

        if not hasattr(request, "basket"):
            return response

        # If the basket was never initialized we can safely return
        if (
            isinstance(request.basket, SimpleLazyObject)
            and request.basket._wrapped is empty
        ):
            return response

        cookie_key = self.get_cookie_key(request)
        # Check if we need to set a cookie. If the cookies is already available
        # but is set in the cookies_to_delete list then we need to re-set it.
        has_basket_cookie = (
            cookie_key in request.COOKIES and cookie_key not in cookies_to_delete
        )

        # If a basket has had products added to it, but the user is anonymous
        # then we need to assign it to a cookie
        if (
            request.basket.id
            and not request.user.is_authenticated
            and not has_basket_cookie
        ):
            cookie = self.get_basket_hash(request.basket.id)
            response.set_cookie(
                cookie_key,
                cookie,
                max_age=settings.OSCAR_BASKET_COOKIE_LIFETIME,
                secure=settings.OSCAR_BASKET_COOKIE_SECURE,
                httponly=True,
            )
        return response

    # pylint: disable=unused-argument
    def get_cookie_key(self, request):
        """
        Returns the cookie name to use for storing a cookie basket.

        The method serves as a useful hook in multi-site scenarios where
        different baskets might be needed.
        """
        return settings.OSCAR_BASKET_COOKIE_OPEN

    def process_template_response(self, request, response):
        if hasattr(response, "context_data"):
            if response.context_data is None:
                response.context_data = {}
            if "basket" not in response.context_data:
                response.context_data["basket"] = request.basket
            else:
                # Occasionally, a view will want to pass an alternative basket
                # to be rendered.  This can happen as part of checkout
                # processes where the submitted basket is frozen when the
                # customer is redirected to another site (eg PayPal).  When the
                # customer returns and we want to show the order preview
                # template, we need to ensure that the frozen basket gets
                # rendered (not request.basket).  We still keep a reference to
                # the request basket (just in case).
                response.context_data["request_basket"] = request.basket
        return response

    # Helper methods

    def get_basket(self, request):
        """
        Return the open basket for this request
        """
        # pylint: disable=protected-access
        if request._basket_cache is not None:
            return request._basket_cache

        num_items_merged = 0
        manager = Basket.open
        cookie_key = self.get_cookie_key(request)
        cookie_basket = self.get_cookie_basket(cookie_key, request, manager)

        if hasattr(request, "user") and request.user.is_authenticated:
            # Signed-in user: if they have a cookie basket too, it means
            # that they have just signed in and we need to merge their cookie
            # basket into their user basket, then delete the cookie.
            try:
                basket, __ = manager.get_or_create(owner=request.user)
            except Basket.MultipleObjectsReturned:
                # Not sure quite how we end up here with multiple baskets.
                # We merge them and create a fresh one
                old_baskets = list(manager.filter(owner=request.user))
                basket = old_baskets[0]
                for other_basket in old_baskets[1:]:
                    self.merge_baskets(basket, other_basket)
                    # count number of items that have been merged
                    num_items_merged += other_basket.num_items

            # Assign user onto basket to prevent further SQL queries when
            # basket.owner is accessed.
            basket.owner = request.user

            if cookie_basket:
                # count number of items in the basket
                num_items_merged += cookie_basket.num_items
                self.merge_baskets(basket, cookie_basket)
                request.cookies_to_delete.append(cookie_key)

        elif cookie_basket:
            # Anonymous user with a basket tied to the cookie
            basket = cookie_basket
        else:
            # Anonymous user with no basket - instantiate a new basket
            # instance.  No need to save yet.
            basket = Basket()

        # Cache basket instance for the during of this request
        request._basket_cache = basket

        if num_items_merged > 0:
            # show warning only if items have been merged
            messages.add_message(
                request,
                messages.WARNING,
                ngettext_lazy(
                    "We have merged %(num_items_merged)d item from a "
                    "previous session to your basket.",
                    "We have merged %(num_items_merged)d items from a "
                    "previous session to your basket.",
                    num_items_merged,
                )
                % {"num_items_merged": num_items_merged},
            )

        return basket

    def merge_baskets(self, master, slave):
        """
        Merge one basket into another.

        This is its own method to allow it to be overridden
        """
        master.merge(slave, add_quantities=False)

    # pylint: disable=unused-argument
    def get_cookie_basket(self, cookie_key, request, manager):
        """
        Looks for a basket which is referenced by a cookie.

        If a cookie key is found with no matching basket, then we add
        it to the list to be deleted.
        """
        basket = None
        if cookie_key in request.COOKIES:
            basket_hash = request.COOKIES[cookie_key]
            try:
                basket_id = Signer().unsign(basket_hash)
                basket = Basket.objects.get(
                    pk=basket_id, owner=None, status=Basket.OPEN
                )
            except (BadSignature, Basket.DoesNotExist):
                request.cookies_to_delete.append(cookie_key)
        return basket

    def apply_offers_to_basket(self, request, basket):
        if not basket.is_empty:
            Applicator().apply(basket, request.user, request)

    def get_basket_hash(self, basket_id):
        return Signer().sign(basket_id)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import oscar.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('offer', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='benefit',
            name='proxy_class',
            field=oscar.models.fields.NullCharField(default=None, max_length=255, verbose_name='Custom class'),
            preserve_default=True,
        ),
    ]



from django.apps import apps
from django.urls import include, path, re_path
from django.utils.translation import gettext_lazy as _

from oscar.core.application import OscarConfig
from oscar.core.loading import get_class


class CatalogueOnlyConfig(OscarConfig):
    label = "catalogue"
    name = "oscar.apps.catalogue"
    verbose_name = _("Catalogue")

    namespace = "catalogue"

    # pylint: disable=attribute-defined-outside-init, unused-import
    def ready(self):
        from . import receivers

        super().ready()

        self.detail_view = get_class("catalogue.views", "ProductDetailView")
        self.catalogue_view = get_class("catalogue.views", "CatalogueView")
        self.category_view = get_class("catalogue.views", "ProductCategoryView")
        self.range_view = get_class("offer.views", "RangeDetailView")

    def get_urls(self):
        urls = super().get_urls()
        urls += [
            path("", self.catalogue_view.as_view(), name="index"),
            re_path(
                r"^(?P<product_slug>[\w-]*)_(?P<pk>\d+)/$",
                self.detail_view.as_view(),
                name="detail",
            ),
            re_path(
                r"^category/(?P<category_slug>[\w-]+(/[\w-]+)*)_(?P<pk>\d+)/$",
                self.category_view.as_view(),
                name="category",
            ),
            re_path(
                r"^ranges/(?P<slug>[\w-]+)/$", self.range_view.as_view(), name="range"
            ),
        ]
        return self.post_process_urls(urls)


class CatalogueReviewsOnlyConfig(OscarConfig):
    label = "catalogue"
    name = "oscar.apps.catalogue"
    verbose_name = _("Catalogue")

    # pylint: disable=attribute-defined-outside-init, unused-import
    def ready(self):
        from . import receivers

        super().ready()

        self.reviews_app = apps.get_app_config("reviews")

    def get_urls(self):
        urls = super().get_urls()
        urls += [
            re_path(
                r"^(?P<product_slug>[\w-]*)_(?P<product_pk>\d+)/reviews/",
                include(self.reviews_app.urls[0]),
            ),
        ]
        return self.post_process_urls(urls)


class CatalogueConfig(CatalogueOnlyConfig, CatalogueReviewsOnlyConfig):
    """
    Composite class combining Products with Reviews
    """



import factory

from oscar.core.loading import get_model

__all__ = [
    "SourceTypeFactory",
    "SourceFactory",
    "TransactionFactory",
]


class SourceTypeFactory(factory.django.DjangoModelFactory):
    name = "Creditcard"
    code = "creditcard"

    class Meta:
        model = get_model("payment", "SourceType")


class SourceFactory(factory.django.DjangoModelFactory):
    order = factory.SubFactory("oscar.test.factories.OrderFactory")
    source_type = factory.SubFactory(SourceTypeFactory)

    class Meta:
        model = get_model("payment", "Source")


class TransactionFactory(factory.django.DjangoModelFactory):
    amount = factory.LazyAttribute(lambda obj: obj.source.order.total_incl_tax)
    reference = factory.LazyAttribute(lambda obj: obj.source.order.number)
    source = factory.SubFactory(SourceFactory)
    status = "authorised"

    class Meta:
        model = get_model("payment", "Transaction")



from django.shortcuts import render


# pylint: disable=unused-argument
def handler403(request, exception):
    return render(request, "oscar/403.html", status=403)


# pylint: disable=unused-argument
def handler404(request, exception):
    return render(request, "oscar/404.html", status=404)


def handler500(request):
    return render(request, "oscar/500.html", status=500)


def sort_queryset(queryset, request, allowed_sorts, default=None):
    """Sorts the queryset by one of allowed_sorts based on parameters
    'sort' and 'dir' from request"""
    sort = request.GET.get("sort", None)
    if sort in allowed_sorts:
        direction = request.GET.get("dir", "asc")
        sort = ("-" if direction == "desc" else "") + sort
        queryset = queryset.order_by(sort)
    elif default:
        queryset = queryset.order_by(default)
    return queryset



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-04-18 20:32
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('voucher', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='voucher',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True),
        ),
        migrations.AlterField(
            model_name='voucherapplication',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True),
        ),
    ]



# -*- coding: utf-8 -*-
from django.forms.models import inlineformset_factory

from oscar.core.loading import get_classes, get_model

WishList = get_model("wishlists", "WishList")
Line = get_model("wishlists", "Line")
WishListSharedEmail = get_model("wishlists", "WishListSharedEmail")

WishListLineForm, WishListSharedEmailForm = get_classes(
    "wishlists.forms", ("WishListLineForm", "WishListSharedEmailForm")
)


LineFormset = inlineformset_factory(
    WishList,
    Line,
    fields=("quantity",),
    form=WishListLineForm,
    extra=0,
    can_delete=False,
)
WishListSharedEmailFormset = inlineformset_factory(
    WishList,
    WishListSharedEmail,
    fields=("email",),
    form=WishListSharedEmailForm,
    extra=3,
    can_delete=True,
)
